
from timeit import default_timer as timer
import warnings
import torch
import torch.nn as nn

# optional imports for tensor decomposition
try:
	import tensorly as tl
	from tensorly.decomposition import non_negative_parafac, parafac
	tl.set_backend('pytorch')
except ImportError:
	warnings.warn("Package 'tensorly' not installed. "
				  "Tensor decomposition is not supported.",
				  category=ImportWarning)
	tl, non_negative_parafac, parafac = None, None, None

from smoother.losses import SpatialLoss

class CPDecomposition(nn.Module):
	"""Tensor CP decomposition."""
	def __init__(self, shape_tensor, dim_hidden, nonneg = True) -> None:
		super().__init__()
		self.cpd_configs = {}
		self.nonneg = nonneg

		self.dim_x, self.dim_y, self.dim_z = shape_tensor
		self.dim_hidden = dim_hidden

		# factorized low-dimensional embedding
		# target = weights * outer(x_hidden, y_hidden, z_hidden)
		self.weights = nn.Parameter(torch.ones(dim_hidden), requires_grad=False)
		self.x_hidden = nn.Parameter(torch.rand(self.dim_x, dim_hidden), requires_grad=True)
		self.y_hidden = nn.Parameter(torch.rand(self.dim_y, dim_hidden), requires_grad=True)
		self.z_hidden = nn.Parameter(torch.rand(self.dim_z, dim_hidden), requires_grad=True)

	def forward(self):
		# target = weights * outer(x_hidden, y_hidden, z_hidden)
		if not self.nonneg:
			return tl.cp_to_tensor(
	  			(self.weights, (self.x_hidden, self.y_hidden, self.z_hidden))
			)
		return tl.cp_to_tensor(
			(self.weights, (torch.abs(self.x_hidden), torch.abs(self.y_hidden), torch.abs(self.z_hidden)))
		)

	def get_loss_fn(self):
		return torch.nn.MSELoss()

	def init_with_tensorly(self, target):
		"""Initialize tensor decomposition using tensorly."""
		if self.nonneg:
			weights, factors = non_negative_parafac(target, self.dim_hidden, init='svd')
		else:
			weights, factors = parafac(target, self.dim_hidden, init='svd')
		self.weights.data = weights
		self.x_hidden.data = factors[0]
		self.y_hidden.data = factors[1]
		self.z_hidden.data = factors[2]

	def decomposition_sp(self, target,
						spatial_loss : SpatialLoss = None, lambda_spatial_loss = 0.1,
						lr = 1e-3, max_epochs = 1000, patience = 10,
						init_with_tensorly = False, verbose = True, quite = False):
		self.cpd_configs = {
			'spatial_loss':spatial_loss, 'lambda_spatial_loss':lambda_spatial_loss,
			'lr':lr, 'max_epochs':max_epochs, 'patience':patience,
			'init_with_tensorly':init_with_tensorly, 'verbose':verbose, 'quite':quite,
			'return_model':False
		}
		return _decomposition_sp(target, model=self, **self.cpd_configs)


def _decomposition_sp(target, model,
		   spatial_loss : SpatialLoss = None, lambda_spatial_loss = 0.1,
		   lr = 1e-3, max_epochs = 1000, patience = 10,
		   init_with_tensorly = True, verbose = True, quite = False, return_model = False):
	# start timer
	t_start = timer()

	# define loss function and optimizer
	loss_fn = model.get_loss_fn()
	optimizer = torch.optim.Adam(model.parameters(), lr=lr)

	# initialize tensor decomposition
	if init_with_tensorly:
		model.init_with_tensorly(target)

	# start training
	model.train()

	# set iteration limits
	epoch, prev_loss, d_loss = 0, 100000.0, 100000.0
	max_epochs = 10000 if max_epochs == -1 else max_epochs
	patience = patience if patience > 0 else 1

	while epoch < max_epochs and d_loss >= 1e-5 or (d_loss < 0 and patience > 0):

		optimizer.zero_grad()

		# reconstruct target from hidden factors
		outputs = model()

		# calculate reconstruction error
		loss = loss_fn(outputs, target) # mse loss

		# store the model loss
		model_loss = loss.detach().item()

		# add spatial loss
		if spatial_loss is not None:
			sp_rv = torch.cat([model.x_hidden, model.y_hidden], dim=1).T
			loss += lambda_spatial_loss * spatial_loss(sp_rv)

		# backpropagate and update weights
		loss.backward()
		optimizer.step()

		# check convergence
		epoch += 1
		d_loss = prev_loss - loss.detach().item()
		if d_loss < 0: # if loss increases
			patience -= 1

		prev_loss = loss.detach().item()

		if (verbose and not quite) and epoch % 10 == 0:
			print(f'Epoch {epoch}. Total loss {prev_loss:.4f}. Model loss {model_loss:.4f}.')

	t_end = timer()

	if not quite: # print final message
		print(f'=== Time {t_end - t_start : .2f}s. Total epoch {epoch}. '
			  f'Final loss {prev_loss:.2f}.')

	if d_loss >= 1e-5:
		warnings.warn("Fail to converge. Try to increase 'max_epochs'.")

	# Depending on the weight matrix, sometimes the inverse covariance matrix can
	# have negative eigenvalues, be careful about the regression coefficient
	if prev_loss < 0:
		warnings.warn(
	  		"Negative loss occurred because the inverse covariance matrix is not positive semidefinite."
			"Try to restrict 'spatial_weights' or use a smaller 'smooth_l'."
		)

	if return_model:
		return model



"""
Impute spatial feature of interest with spatial loss
"""
import warnings
from timeit import default_timer as timer
from scipy.sparse import coo_matrix
import torch
import torch.nn as nn
from smoother.losses import SpatialLoss

# optional imports for convex optimization
try:
	import cvxpy as cp
	from cvxpy.atoms.affine.wraps import psd_wrap
except ImportError:
	warnings.warn("Package 'cvxpy' not installed so that "
				  "Convex optimization solvers are not supported.",
				  category=ImportWarning)
	cp, psd_wrap = None, None

class ImputeModel():
	"""Class for different implementations of spatial imputation.

	This class is designed for coorperative inheritance (mixin).

	Attributes:
		y_obs (2D tensor): Observed spatial feature matrix, n_obs x n_feature.
		spatial_loss_all (SpatialLoss): Spatial loss object build from the combined
			spatial coordinates containing both observed (the first n_obs rows)
			and missing spots (the rest rows).
		fixed_obs (bool): Whether to allow updates in the observed data.
		nonneg (bool): Whether to enforce nonnegativity on spatial features.
		n_feature (int): Number of spatial features to process.
		n_all (int): Number of total spots. n_all = n_obs + n_missing.
		n_obs (int): Number of observed spots.
		n_missing (int): Number of missing spots.
		var_obs : Observed spatial feature matrix, n_obs x n_feature.
		var_missing : Missing spatial feature matrix to impute, n_missing x n_feature.
		var_all : Concatenated observed and missing spatial feature matrix, n_all x n_feature.
		impute_configs (dict): Dictionary of imputation configurations.
		impute_time (float): Time spent on deconvolution.
		impute_flag (bool): Whether the imputation was successful.
	"""
	def __init__(self, y_obs, spatial_loss_all : SpatialLoss,
				 fixed_obs = True, nonneg = False, **kwargs):
		"""Initialize ImputeModel object."""
		# for the ImputeModel class to be compatible with nn.module,
		# here we need to forward all unused arguments
		super().__init__(**kwargs)

		self.y_obs = None
		self.spatial_loss_all = None
		self.fixed_obs = None # whether to allow updates in the observed data
		self.nonneg = None # whether to enforce nonnegativity on spatial features

		self.n_feature = None # number of spatial features to process
		# the first n_obs spots modelled in spatial_loss are the observed data
		# and the rest are the missing data to impute
		self.n_obs = None # number of observed spots
		self.n_all = None # number of total spots
		self.n_missing = None # number of missing spots

		# set model parameters
		self.var_obs = None # (n_obs, n_feature)
		self.var_missing = None # (n_missing, n_feature)
		self.var_all = None # (n_all, n_feature)
		self.set_params(y_obs, spatial_loss_all, fixed_obs, nonneg)

		# imputation attributes
		self.impute_configs = None
		self.impute_time = 0

	def set_params(self, y_obs, spatial_loss_all, fixed_obs = True, nonneg = False):
		"""Set configuration parameters and initialize variables."""
		# currently only support one covariance structure
		if len(spatial_loss_all.inv_cov) > 1:
			raise NotImplementedError

		# store raw data and spatial loss
		self.y_obs = y_obs
		self.spatial_loss_all = spatial_loss_all
		self.fixed_obs = fixed_obs # whether to allow updates in the observed data

		# sanity check for the nonneg argument
		if nonneg and self.y_obs.min() < 0:
			raise ValueError("Nonnegativity is enforced but the observed data contains negative values.")
		self.nonneg = nonneg # whether to enforce nonnegativity on spatial features

		self.n_feature = y_obs.shape[1] # number of spatial features to process
		# the first n_obs spots are the observed data
		# and the rest are the missing data to impute
		self.n_obs = y_obs.shape[0] # number of observed spots
		self.n_all = spatial_loss_all.inv_cov[0].shape[0] # number of total spots
		self.n_missing = self.n_all - self.n_obs # number of missing spots

		# initialze variables to impute()
		self.set_variables()

	def set_variables(self):
		"""Set variables to impute."""
		raise NotImplementedError

	def get_recon_loss(self):
		"""Get reconstruction loss."""
		raise NotImplementedError

	def get_sp_loss(self):
		"""Get spatial loss (weighted by lambda_spatial_loss)."""
		raise NotImplementedError

	def get_results(self):
		"""Get spatial features after imputation."""
		raise NotImplementedError


class ImputeTorch(ImputeModel, nn.Module):
	"""Imputation solver implemented using pytorch.

	Attributes:
		var_obs (2D tensor): Observed spatial feature matrix, n_obs x n_feature.
		var_missing (2D tensor): Missing spatial feature matrix to impute, n_missing x n_feature.
		var_all (2D tensor): Concatenated observed and missing spatial feature matrix, n_all x n_feature.
		See super class attributes for additional arguments.
	"""
	def __init__(self, y_obs, spatial_loss_all : SpatialLoss,
				 fixed_obs = True, nonneg = False, **kwargs) -> None:
		"""Initialize and run the imputation.

		Args:
			y_obs (2D tensor): Observed spatial feature matrix, n_obs x n_feature.
			spatial_loss_all (SpatialLoss): Spatial loss object build from the combined
						spatial coordinates containing both observed (the first n_obs rows)
						and missing spots (the rest rows).
			See class attributes for additional arguments.
		"""
		super().__init__(y_obs, spatial_loss_all, fixed_obs, nonneg)

		# run imputation
		self.impute_flag = self.impute(**kwargs)

	def set_variables(self):
		"""Initialize torch parameters to estimate."""
		# set torch variables
		self.var_obs = nn.parameter.Parameter(self.y_obs.clone(), requires_grad = (not self.fixed_obs))
		self.var_missing = nn.parameter.Parameter(
			torch.ones(self.n_missing, self.n_feature) * self.y_obs.mean(dim=0, keepdim=True),
			requires_grad = True)
		self.var_all = torch.concat([self.var_obs, self.var_missing], dim = 0)

	def forward(self, y_obs, **kwargs):
		"""Run imputation for a new set of spatial features."""
		self.set_params(y_obs, self.spatial_loss_all, self.fixed_obs, self.nonneg)
		self.impute_flag = self.impute(**kwargs)
		return self.var_all.detach()

	def get_recon_loss(self):
		"""Calculate loss of deviation from observation.

		If the observed data is not fixed, then the loss is used to make sure data does not deviate
		too much from the observed value after imputation.
		"""
		if self.fixed_obs:
			return 0
		else:
			loss_fn = torch.nn.MSELoss() # mean squared error
			if self.nonneg:
				return loss_fn(self.y_obs, self.var_obs.abs())
			else:
				return loss_fn(self.y_obs, self.var_obs)

	def get_sp_loss(self):
		"""Calculate loss of spatial smoothness."""
		l_sp_loss = self.impute_configs.get('lambda_spatial_loss')
		sp_loss = self.impute_configs.get('spatial_loss')
		# calculate spatial loss
		if self.nonneg:
			return l_sp_loss * sp_loss(self.var_all.abs().T)
		else:
			return l_sp_loss * sp_loss(self.var_all.T)

	def get_results(self):
		"""Get spatial features of all spots after imputation.

		The first n_obs spots are the observed spots and the rest are the missing ones.
		If fixed_obs is True, then the returned observed data is not updated.
		"""
		return self.var_all.detach()

	def final_sanity_check(self):
		if self.nonneg:
			with torch.no_grad():
				self.var_obs.copy_(self.var_obs.abs())
				self.var_missing.copy_(self.var_missing.abs())
				self.var_all.copy_(self.var_all.abs())

	def impute(self, lambda_spatial_loss = 1.0, lr = 1e-3, max_epochs = 1000,
			   patience = 10, tol = 1e-5, verbose = True, quiet = False) -> bool:
		"""Run imputation algorithm.

		Args:
			lambda_spatial_loss (float): Specifies the strength of the spatial smoothing.
			lr (float): Learning rate.
			max_epochs (int): Maximum number of training epochs. If -1, iterate until
				convergence (`d_loss` < tol).
			patient (int): Number of epochs to wait for the loss to decrease before stopping.
			tol (float): Tolerance of loss convergence.
			verbose (bool): If True, print out loss while training.
			quiet (bool): If True, no output printed.

		Returns:
			bool: True if imputation is successful.
		"""
		self.impute_configs = {
			'y_obs':self.y_obs, 'spatial_loss':self.spatial_loss_all,
			'n_feature':self.n_feature, 'n_obs':self.n_obs, 'n_missing':self.n_missing,
			'fixed_obs':self.fixed_obs, 'nonneg':self.nonneg,
			'lambda_spatial_loss':lambda_spatial_loss,
			'lr':lr, 'max_epochs':max_epochs, 'patience':patience, 'tol':tol,
			'verbose':verbose, 'quiet':quiet
		}

		if self.n_missing == 0 and self.fixed_obs:
			warnings.warn("Imputation stopped: Observation is fixed with no missing position.")
			return False

		# start timer
		t_start = timer()

		# initialize optimizer
		optimizer = torch.optim.Adam(self.parameters(), lr=lr)

		# start training
		self.train()

		# set iteration limits
		epoch, prev_loss, d_loss = 0, 100000.0, 100000.0
		max_epochs = 10000 if max_epochs == -1 else max_epochs
		patience = patience if patience > 0 else 1

		while epoch < max_epochs and (d_loss >= tol or (d_loss < 0 and patience > 0)):
			optimizer.zero_grad()

			# calculate reconstruction loss
			loss = self.get_recon_loss()

			# store the reconstruction loss
			if self.fixed_obs:
				recon_loss = 0.0
			else:
				recon_loss = loss.detach().item()

			# add spatial loss
			sp_loss = self.get_sp_loss()
			loss += sp_loss
			sp_loss = sp_loss.detach().item()

			# backpropagate and update weights
			loss.backward()
			optimizer.step()

			# update the merged vector
			self.var_all = torch.concat([self.var_obs, self.var_missing], dim = 0)

			# check convergence
			epoch += 1
			d_loss = prev_loss - loss.detach().item()
			if d_loss < 0: # if loss increases
				patience -= 1

			# print out loss
			prev_loss = loss.detach().item()

			if (verbose and not quiet) and epoch % 10 == 0:
				print(f'Epoch {epoch}. Loss: (total) {prev_loss:.4f}. '
					f'(recon) {recon_loss:.4f}. (spatial) {sp_loss:.4f}.')

		# make sure constraints are satisfied
		self.final_sanity_check()

		t_end = timer()

		# save runtime
		self.impute_time = t_end - t_start

		if not quiet: # print final message
			print(f'=== Time {self.impute_time : .2f}s. Total epoch {epoch}. '
				f'Final loss: (total) {prev_loss : .3f}. (spatial) {sp_loss : .3f}.')

		# if d_loss >= tol:
		if epoch == max_epochs:
			warnings.warn("Fail to converge. Try to increase 'max_epochs'.")
			return False

		# Depending on the weight matrix, sometimes the inverse covariance matrix can
		# have negative eigenvalues, be careful about the regression coefficient
		if prev_loss < 0:
			warnings.warn(
				"Negative loss occurred because the inverse covariance matrix is not positive semidefinite. "
				"Try to restrict 'spatial_weights' or use a smaller 'smooth_l'."
			)
			return False

		return True


class ImputeConvex(ImputeModel):
	"""Imputation solver implemented using cvxpy.

	Attributes:
		var_obs : Observed spatial feature matrix, n_obs x n_feature.
		var_missing : Missing spatial feature matrix to impute, n_missing x n_feature.
		var_all : Concatenated observed and missing spatial feature matrix, n_all x n_feature.
		recon_loss_exp (cp.Expression): Reconstruction loss expression.
		sp_loss_exp (cp.Expression): Spatial loss expression.
		See super class attributes for additional arguments.
	"""
	def __init__(self, y_obs, spatial_loss_all : SpatialLoss,
				 fixed_obs = True, nonneg = False, **kwargs) -> None:
		"""Initialize and run the imputation.

		Args:
			y_obs (2D tensor): Observed spatial feature matrix, n_obs x n_feature.
			spatial_loss_all (SpatialLoss): Spatial loss object build from the combined
						spatial coordinates containing both observed (the first n_obs rows)
						and missing spots (the rest rows).
			See class attributes for additional arguments.
		"""
		self.recon_loss_exp = None # cvxpy expression for reconstruction loss
		self.spatial_loss_exp = None # cvxpy expression for spatial loss
		super().__init__(y_obs, spatial_loss_all, fixed_obs, nonneg)

		# run imputation
		self.impute_flag = self.impute(**kwargs)

	def set_variables(self):
		"""Initialize parameters to estimate."""
		# set cvxpy variables
		if not self.fixed_obs:
			self.var_all = cp.Variable((self.n_all, self.n_feature), nonneg=self.nonneg)
			self.var_obs = self.var_all[0:self.n_obs, :]
			self.var_missing = self.var_all[self.n_obs:, :]
		else:
			self.var_missing = cp.Variable((self.n_missing, self.n_feature), nonneg=self.nonneg)
			self.var_obs = self.y_obs
			self.var_all = cp.vstack([self.var_obs, self.var_missing])

		# set cvxpy expressions
		self.set_recon_loss()
		self.set_sp_loss()

	def set_recon_loss(self):
		"""Set loss expression of deviation from observation.

		If the observed data is not fixed, then the loss is used to make sure data does not deviate
		too much from the observed value after imputation.
		"""
		if self.fixed_obs:
			self.recon_loss_exp = cp.Constant(0.0)
		else:
			self.recon_loss_exp = cp.sum_squares(self.var_obs - self.y_obs) / (self.n_obs * self.n_feature)

	def set_sp_loss(self):
		"""Set spatial loss expression (before weighted by lambda_spatial_loss)."""
		sp_loss = self.spatial_loss_all
		self.spatial_loss_exp = 0

		# extract inverse covariance matrix
		if not sp_loss.standardize_cov:
			raise ValueError("Spatial covariance must be standardized.")
		inv_cov = sp_loss.inv_cov[0].coalesce()

		# use scipy sparse matrix to speed up computation
		if sp_loss.use_sparse:
			inv_cov = psd_wrap(coo_matrix(
				(inv_cov.values().numpy(), inv_cov.indices().numpy()), inv_cov.shape
			))

		for i in range(self.n_feature):
			self.spatial_loss_exp += \
				cp.quad_form(self.var_all[:,i], inv_cov)

		self.spatial_loss_exp /= (self.n_all * self.n_feature)

	def get_recon_loss(self):
		"""Calculate loss of deviation from observation.

		If the observed data is not fixed, then the loss is used to make sure data does not deviate
		too much from the observed value after imputation.
		"""
		return self.recon_loss_exp

	def get_sp_loss(self):
		"""Calculate loss of spatial smoothness (weighted by lambda_spatial_loss)."""
		l_sp_loss = self.impute_configs.get('lambda_spatial_loss')
		return self.spatial_loss_exp * l_sp_loss

	def get_results(self):
		"""Get spatial features of all spots after imputation.

		The first n_obs spots are the observed spots and the rest are the missing ones.
		If fixed_obs is True, then the returned observed data is not updated.
		"""
		return self.var_all.value

	def impute(self, lambda_spatial_loss = 1.0,
			   verbose = False, quiet = False, solver = None, **kwargs) -> bool:
		"""Run imputation algorithm.

		Args:
			lambda_spatial_loss (float): Specifies the strength of the spatial smoothing.
			verbose (bool): If True, print out loss while training.
			quiet (bool): If True, no output printed.

		Returns:
			bool: True if imputation is successful.
		"""
		self.impute_configs = {
			'y_obs':self.y_obs, 'spatial_loss':self.spatial_loss_all,
			'n_feature':self.n_feature, 'n_obs':self.n_obs, 'n_missing':self.n_missing,
			'fixed_obs':self.fixed_obs, 'nonneg':self.nonneg,
			'lambda_spatial_loss':lambda_spatial_loss,
			'verbose':verbose, 'quiet':quiet
		}

		if self.n_missing == 0 and self.fixed_obs:
			warnings.warn("Imputation stopped: Observation is fixed with no missing position.")
			return False

		# start timer
		t_start = timer()

		# initialize optimization problem
		self.obj = cp.Minimize(self.get_recon_loss() + self.get_sp_loss())
		self.problem = cp.Problem(self.obj)

		# pass solver arguments
		print_sol_out = verbose and (not quiet)
		self.problem.solve(solver = solver, verbose = print_sol_out, **kwargs)

		# check problem status
		if not self.problem.status == cp.OPTIMAL:
			if self.problem.status == cp.UNBOUNDED:
				raise ValueError('Solver encountered numeric issues. '
								 'Try torch-based models or reduce number of genes.')
			warnings.warn(f"Problem status: {self.problem.status}.")

		# extract losses
		total_loss = self.obj.value
		recon_loss = self.get_recon_loss().value.item()
		sp_loss = self.get_sp_loss().value.item()

		t_end = timer()

		# save runtime
		self.impute_time = t_end - t_start

		if not quiet: # print final message
			print(f'=== Time {self.impute_time : .2f}s. '
				f'Loss: (total) {total_loss : .3f}, '
				f'(recon) {recon_loss : .3f}, '
				f'(spatial) {sp_loss : .3f}')
		return True



"""
Deconvolute spatial-omics data with spatial loss
"""
import warnings
from functools import partial
from timeit import default_timer as timer
import numpy as np
from scipy.sparse import coo_matrix
import torch
import torch.nn as nn
from torch.autograd import Variable
from smoother.losses import SpatialLoss, TopLoss

# optional imports for convex optimization
try:
	import cvxpy as cp
	from cvxpy.atoms.affine.wraps import psd_wrap
except ImportError:
	warnings.warn("Package 'cvxpy' not installed so that "
				  "Convex optimization solvers are not supported.",
				  category=ImportWarning)
	cp, psd_wrap = None, None

torch.set_default_dtype(torch.float32)

class DeconvModel():
	"""Interface for convex deconvolution models."""
	def get_model_loss(self):
		"""Get regression reconstruction loss."""
		raise NotImplementedError

	def get_sp_loss(self):
		"""Get spatial loss (weighted by lambda_spatial_loss)."""
		raise NotImplementedError

	def get_props(self):
		"""Get the predicted celltype abundance from the trained deconvolution model."""
		raise NotImplementedError

def LinearRegression(backend = 'pytorch', **kwargs) -> DeconvModel:
	assert backend in ['pytorch', 'cvxpy']
	if backend == 'pytorch':
		return LinearRegressionTorch(**kwargs)
	else:
		return LinearRegressionConvex(**kwargs)

def NNLS(backend = 'pytorch', **kwargs) -> DeconvModel:
	assert backend in ['pytorch', 'cvxpy']
	if backend == 'pytorch':
		return NNLSTorch(**kwargs)
	else:
		return NNLSConvex(**kwargs)

def NuSVR(backend = 'pytorch', **kwargs) -> DeconvModel:
	assert backend in ['pytorch', 'cvxpy']
	if backend == 'pytorch':
		return NuSVRTorch(**kwargs)
	else:
		return NuSVRConvex(**kwargs)

def DWLS(backend = 'pytorch', **kwargs) -> DeconvModel:
	assert backend in ['pytorch', 'cvxpy']
	if backend == 'pytorch':
		return DWLSTorch(**kwargs)
	else:
		return DWLSConvex(**kwargs)

def LogNormReg(backend = 'pytorch', **kwargs) -> DeconvModel:
	assert backend in ['pytorch']
	if backend == 'pytorch':
		return LogNormRegTorch(**kwargs)

class DeconvModelTorch(nn.Module, DeconvModel):
	"""Class for deconvolution models implemented using pytorch.

	Attributes:
		dim_in (int): Input dimension, also number of groups.
		dim_out (int): Output dimension, also number of spots.
		use_bias (bool): Whether to add bias term in regression.
		model_name (str): Name of the deconvolution model.
		nonneg (bool): Whether to apply nonnegative contraint
			on spatial variables.
		deconv_configs (dict): Dictionary of deconvolution configurations.
		deconv_time (float): Time spent on deconvolution.
	"""
	def __init__(self, dim_in, dim_out, model_name) -> None:
		super().__init__()
		self.dim_in = dim_in
		self.dim_out = dim_out
		self.model_name = model_name
		self.use_bias = None
		self.nonneg = None # nonnegative constraint
		self.deconv_configs = None # deconvolution configs
		self.deconv_time = None # time spent in deconvolution

	def set_params(self, dim_in, dim_out, bias):
		raise NotImplementedError

	def get_top_loss(self):
		"""Get toplogical loss (weighted by lambda_top_loss)."""
		raise NotImplementedError

	def init_with_lr_sol(self, x, y):
		"""Initialize model with linear regression solution."""
		raise NotImplementedError

	def final_sanity_check(self):
		"""Make sure constraints are followed after deconvolution."""
		raise NotImplementedError

	def deconv(self, x, y,
			   spatial_loss : SpatialLoss = None, lambda_spatial_loss = 0.0,
			   top_loss : TopLoss = None, lambda_top_loss = 0.0,
			   lr = 1e-3, max_epochs = 1000, patience = 10, tol = 1e-5,
			   init_with_lr_sol = True, verbose = True, quiet = False):
		"""Deconvolute spatial-omics data with spatial loss.

		Loss_total = Loss_model + `lambda_spatial_loss` * Loss_spatial +
			`lambda_top_loss` * Loss_topological.

		Args:
			x (2D tensor): Bulk feature signiture matrix, num_feature x num_group.
			y (2D tensor): Spatial feature matrix, num_feature x num_spot.
			spatial_loss (SpatialLoss): The spatial smoothing loss.
				Spatial prior can be one of 'none', 'kl', 'sma', 'sar', 'car', 'icar'
				- KL : KL divergence of cell type proportion vectors of neiboring spots
				- SMA : Spatial moving average
				- SAR : Simultaneous auto-regressive model
				- CAR : Conditional auto-regressive model
				- ICAR : Intrinsic conditional auto-regressive model
				See model description for more details.
			lambda_spatial_loss (float): Specifies the strength of the spatial smoothing.
			top_loss (TopLoss): The topological loss (expected betti number for each group).
				top_loss.betti_priors = {
					group_id : {betti_k : expected number of barcodes (prior)}, ...
				}
			lambda_top_loss (float): Specifies the strength of the topological constraints.
			lr (float): Learning rate.
			max_epochs (int): Maximum number of training epochs. If -1, iterate until
				convergence (`d_loss` < 1e-5).
			patient (int): Number of epochs to wait for the loss to decrease before stopping.
			tol (float): Tolerance of loss convergence.
			init_with_lr_sol (bool): Whether to initialize regression weights with the OLE solution.
			verbose (bool): If True, print out loss while training.
			quiet (bool): If True, no output printed.
		"""
		self.deconv_configs = {
			'x':x, 'y':y,
	  		'spatial_loss':spatial_loss, 'lambda_spatial_loss':lambda_spatial_loss,
			'top_loss':top_loss, 'lambda_top_loss':lambda_top_loss,
			'lr':lr, 'max_epochs':max_epochs, 'patience':patience, 'tol':tol,
			'init_with_lr_sol':init_with_lr_sol, 'verbose':verbose, 'quiet':quiet,
			'return_model':False
		}
		return _deconv_torch(deconv_model=self, **self.deconv_configs)


def _deconv_torch(x, y, deconv_model : DeconvModelTorch,
		   spatial_loss : SpatialLoss = None, lambda_spatial_loss = 1.0,
		   top_loss : TopLoss = None, lambda_top_loss = 0.0,
		   lr = 1e-3, max_epochs = 1000, patience = 10, tol = 1e-5,
		   init_with_lr_sol = True, verbose = True, quiet = False, return_model = True):
	"""Train a regression-based spatial deconvolution model.

	Loss_total = Loss_model + `lambda_spatial_loss` * Loss_spatial +
		`lambda_top_loss` * Loss_topological.

	Args:
		x (2D tensor): Bulk feature signiture matrix, num_feature x num_group.
		y (2D tensor): Spatial feature matrix, num_feature x num_spot.
		deconv_model (DeconvModelTorch): A deconvolution model to be used.
			Model can be one of 'linear', 'nnls', 'svr', or 'dwls'.
			The 'svr' model takes 'C' and 'nu' as two additional inputs, see
			https://scikit-learn.org/stable/modules/generated/sklearn.svm.NuSVR.html.
		spatial_loss (SpatialLoss): The spatial smoothing loss.
			Spatial prior can be one of 'none', 'kl', 'sma', 'sar', 'car', 'icar'
			- KL : KL divergence of cell type proportion vectors of neiboring spots
			- SMA : Spatial moving average
			- SAR : Simultaneous auto-regressive model
			- CAR : Conditional auto-regressive model
			- ICAR : Intrinsic conditional auto-regressive model
			See model description for more details.
		lambda_spatial_loss (float): Specifies the strength of the spatial smoothing.
		top_loss (TopLoss): The topological loss (expected betti number for each group).
			top_loss.betti_priors = {
				group_id : {betti_k : expected number of barcodes (prior)}, ...
			}
		lambda_top_loss (float): Specifies the strength of the topological constraints.
		lr (float): Learning rate.
		max_epochs (int): Maximum number of training epochs. If -1, iterate until
  			convergence (`d_loss` < 1e-5).
		patient (int): Number of epochs to wait for the loss to decrease before stopping.
		tol (float): Tolerance of loss convergence.
		init_with_lr_sol (bool): Whether to initialize regression weights with the OLE solution.
		verbose (bool): If True, print out loss while training.
		quiet (bool): If True, no output printed.
		return_model (bool): If True, return the trained model.
	"""

	# check model dimensions
	dim_in, dim_out = x.shape[1], y.shape[1]
	if (deconv_model.dim_in, deconv_model.dim_out) != (dim_in, dim_out):
		deconv_model.set_params(dim_in, dim_out, deconv_model.use_bias)

	# start timer
	t_start = timer()

	# initialize optimizer
	optimizer = torch.optim.Adam(deconv_model.parameters(), lr=lr)

	# initialize with regression solution (beta = (x^T*x)^-1 * x^T * y)
	if init_with_lr_sol:
		deconv_model.init_with_lr_sol(x, y)

	# start training
	deconv_model.train()

	# set iteration limits
	epoch, prev_loss, d_loss = 0, 100000.0, 100000.0
	max_epochs = 10000 if max_epochs == -1 else max_epochs
	patience = patience if patience > 0 else 1

	while epoch < max_epochs and (d_loss >= tol or (d_loss < 0 and patience > 0)):

		optimizer.zero_grad()

		# prediction of y from linear regression
		# outputs = deconv_model(x)
		# calculate model loss
		loss = deconv_model.get_model_loss()

		# store the model loss
		model_loss = loss.detach().item()

		# add spatial loss
		if spatial_loss is not None and lambda_spatial_loss > 0:
			# sp_loss = lambda_spatial_loss * spatial_loss(deconv_model.lr.weight.T)
			sp_loss = deconv_model.get_sp_loss()
			loss += sp_loss
			sp_loss = sp_loss.detach().item()
		else:
			sp_loss = 0.0

		# add topological loss
		if top_loss is not None and lambda_top_loss > 0:
			# loss += lambda_top_loss * top_loss(deconv_model.lr.weight.T)
			loss += deconv_model.get_top_loss()

		# backpropagate and update weights
		loss.backward()
		optimizer.step()

		# check convergence
		epoch += 1
		d_loss = prev_loss - loss.detach().item()
		if d_loss < 0: # if loss increases
			patience -= 1

		prev_loss = loss.detach().item()

		if (verbose and not quiet) and epoch % 10 == 0:
			print(f'Epoch {epoch}. Loss: (total) {prev_loss:.4f}. '
		 		  f'(model) {model_loss:.4f}. (spatial) {sp_loss:.4f}.')

	# make sure constraints are satisfied
	deconv_model.final_sanity_check()

	t_end = timer()

	# save runtime
	deconv_model.deconv_time = t_end - t_start

	if not quiet: # print final message
		print(f'=== Time {deconv_model.deconv_time : .2f}s. Total epoch {epoch}. '
			  f'Final loss: (total) {prev_loss:.3f}. (spatial) {sp_loss:.3f}.')

	if d_loss >= tol:
		warnings.warn("Fail to converge. Try to increase 'max_epochs'.")

	# Depending on the weight matrix, sometimes the inverse covariance matrix can
	# have negative eigenvalues, be careful about the regression coefficient
	if prev_loss < 0:
		warnings.warn(
			"Negative loss occurred because the inverse covariance matrix is not positive semidefinite. "
			"Try to restrict 'spatial_weights' or use a smaller 'smooth_l'."
		)

	if return_model:
		return deconv_model


class LinearRegressionTorch(DeconvModelTorch):
	"""Linear regression implemented using pytorch."""
	def __init__(self, bias = True, dim_in = 1, dim_out = 1):
		super().__init__(dim_in, dim_out, model_name = 'linear_torch')

		# initialize the linear layer
		self.lr = None
		self.nonneg = False
		self.use_bias = bias
		self.set_params(dim_in, dim_out, bias)

	def set_params(self, dim_in, dim_out, bias = True) -> None:
		self.dim_in = dim_in
		self.dim_out = dim_out
		self.use_bias = bias
		self.lr = nn.Linear(dim_in, dim_out, bias=bias)

	def forward(self, x):
		if self.nonneg: # update with abs(weights)
			if self.use_bias:
				return torch.matmul(x, torch.abs(self.lr.weight.T)) + self.lr.bias
			else:
				return torch.matmul(x, torch.abs(self.lr.weight.T))
		return self.lr(x)

	def get_model_loss_fn(self):
		return torch.nn.MSELoss()

	def get_model_loss(self):
		"""Calculate regression loss."""
		loss_fn = self.get_model_loss_fn() # MSE loss
		# extract deconv inputs
		x, y = self.deconv_configs.get('x'), self.deconv_configs.get('y')
		# self(x) is the model prediction
		return loss_fn(self(x), y)

	def get_sp_loss(self):
		"""Calculate spatial loss (weighted by lambda_spatial_loss)."""
		# extract spatial loss parameters
		l_sp_loss = self.deconv_configs.get('lambda_spatial_loss')
		sp_loss = self.deconv_configs.get('spatial_loss')
		# calculate spatial loss
		if self.nonneg:
			return l_sp_loss * sp_loss(self.lr.weight.T.abs())
		else:
			return l_sp_loss * sp_loss(self.lr.weight.T)

	def get_top_loss(self):
		"""Calculate topological loss (weighted by lambda_top_loss)."""
		# extract topological loss parameters
		l_top_loss = self.deconv_configs.get('lambda_top_loss')
		top_loss = self.deconv_configs.get('top_loss')
		# calculate topological loss
		return l_top_loss * top_loss(self.lr.weight.T)

	def get_props(self):
		"""Get the predicted celltype abundance from the trained deconvolution model."""
		# remove negative abundances
		props = self.lr.weight.data.clamp(min=0) # num_out x num_in
		# normalize
		props = props / (props.sum(1, keepdim=True) + 1e-8)
		return props

	def init_with_lr_sol(self, x, y):
		"""Initialize model with linear regression solution."""
		init_weights = torch.linalg.lstsq(x, y)[0]
		self.lr.weight.data = init_weights.T

	def final_sanity_check(self):
		"""Make sure constraints are satisfied."""
		# for models with non-negative contraint, ensure the final weight is non-negative
		if self.nonneg:
			with torch.no_grad():
				self.lr.weight.copy_(torch.abs(self.lr.weight.data))


class NNLSTorch(LinearRegressionTorch):
	"""Non-negative least square implemented using pytorch."""
	def __init__(self, bias = True, dim_in = 1, dim_out = 1):
		super().__init__(bias, dim_in, dim_out)
		self.model_name = 'nnls_torch'
		self.nonneg = True


class NuSVRTorch(LinearRegressionTorch):
	"""Nu-Support vector regression implemented using pytorch.

	Attributes:
		C (float): SVR regularization parameter.
		nu (float): Nu-SVR parameter.
		epsilon (float): Epsilon in the epsilon-SVR model, specifying the the epsilon-tube within which
			no penalty is associated in the training loss function with points predicted within a distance
			epsilon from the actual value.
		loss_mode (str): Specifies the regression loss function to use. Either 'l2' or 'l1'.
	"""
	def __init__(self, bias = True, dim_in = 1, dim_out = 1,
				 C = 1.0, nu = 0.1, nonneg = False, loss_mode = 'l2'):
		"""Initialize a Nu-SVR deconvolution model.

		Args:
			C (float): SVR regularization parameter.
			nu (float): Nu-SVR parameter.
			nonneg (bool): Whether to require nonnegative regression coefficients.
			loss_mode (str): Specifies the regression loss function to use. Either 'l2' or 'l1'.
		"""
		super().__init__(bias, dim_in, dim_out)

		# store model configs
		self.model_name = 'svr_torch'
		self.nonneg = nonneg
		self.loss_mode = loss_mode
		self.C = Variable(torch.tensor([C]), requires_grad=False)
		self.nu = Variable(torch.tensor([nu]), requires_grad=False)
		self.epsilon = nn.parameter.Parameter(torch.tensor([1.0]), requires_grad=True)

	def get_model_loss_fn(self):
		return partial(_nu_svr_loss, coefs = self.lr.weight.T, epsilon = torch.abs(self.epsilon),
						C = self.C, nu = self.nu, loss = self.loss_mode)

	def final_sanity_check(self):
		"""Make sure constraints are satisfied."""
		super().final_sanity_check()
		# for svr, ensure the final epsilon is always non-neg
		with torch.no_grad():
			self.epsilon.copy_(torch.abs(self.epsilon.data))

def _nu_svr_loss(input, target, coefs, C, nu, epsilon, reduction = 'mean', loss = 'l2'):
	"""Calculate nu-SVR loss.

	Args:
		input (2D tensor): Predicted spatial features (pred = X * beta), num_feature x num_spot.
		target (2D tensor): Observed spatial features, num_feature x num_spot.
		coefs (2D tenor): Columns of regression coefficients, num_group x num_spot.
		C (float): SVR regularization parameter.
		nu (float): Nu-SVR parameter.
		epsilon (float): Epsilon in the epsilon-SVR model, specifying the the epsilon-tube within which
			no penalty is associated in the training loss function with points predicted within a distance
			epsilon from the actual value.
		reduction (str): Specifies the reduction to apply to the output.
		loss (str): Specifies the regression loss function to use. Either 'l2' or 'l1'.

	Return:
		loss (float): Aggregated nu-SVR loss.
	"""
	# calculate regression loss
	if reduction != 'mean':
		raise NotImplementedError("Currently only support mean reduction")

	if loss not in ['l2', 'l1']:
		raise NotImplementedError("Currently only support l2 and l1 loss")

	regression_loss = torch.clamp(torch.abs(target - input) - epsilon, min=0)
	if loss == 'l2':
		regression_loss = regression_loss.pow(2).mean()
	else:
		regression_loss = regression_loss.mean()

	# add l2 loss and penalty on large epsilon
	loss = C * (nu * epsilon + regression_loss) + torch.mean(coefs ** 2) / 2

	return loss


class DWLSTorch(LinearRegressionTorch):
	"""Damped weighted least square implemented using pytorch.

	Here the error term for each feature is scaled by its observed value, not the predicted
	value as described in https://www.nature.com/articles/s41467-019-10802-z.

	Attributes:
		max_weights (float): The upper limit of weights.
	"""
	def __init__(self, bias = True, dim_in = 1, dim_out = 1, nonneg = True, max_weights = 8) -> None:
		super().__init__(bias, dim_in, dim_out)

		# store model configs
		self.model_name = 'dwls_torch'
		self.nonneg = nonneg
		self.max_weights = max_weights

	def get_model_loss_fn(self):
		return partial(_dwls_loss, max_weights = self.max_weights)

def _dwls_loss(input, target, max_weights = 8):
	"""Calculate dampened weighted least square loss.

	Here the error term for each feature is scaled by its observed value, not the predicted
 	value as described in https://www.nature.com/articles/s41467-019-10802-z.

	Args:
		input (2D tensor): Predicted spatial features (pred = X * beta), num_feature x num_spot.
		target (2D tensor): Observed spatial features, num_feature x num_spot.
		max_weights (float): The upper limit of weights.
	"""
	# calculate ordinary least square
	ols = torch.nn.MSELoss(reduction='none')(input, target) # num_feature x num_spot

	# calculate weights for each gene at each spot
	weights = (target ** (-2)).clamp(max = max_weights ** 2) # num_feature x num_spot
	min_weights, _ = torch.min(weights, dim=0, keepdim=True)
	weights = (weights / min_weights).clamp(max = max_weights)

	# calculate weighted least square
	return torch.mul(ols, weights).mean()


class LogNormRegTorch(LinearRegressionTorch):
	"""Log-normal deconvolution implemented using pytorch.

	Minimize MSE(log(Y), log(X @ W)). Here Y and X are all in the raw count space.

	Instead of minimizing the least square loss, this model will minimize MSE after
	log-transforming both observation and prediction, as described in
	in https://www.nature.com/articles/s41467-022-28020-5 as Algorithm 1.

	Attributes:
		epsilon (float): pseudo count added before log transformation.
	"""
	def __init__(self, bias = True, epsilon = 1, dim_in = 1, dim_out = 1):
		super().__init__(bias, dim_in, dim_out)
		self.model_name = 'lognormreg_torch'
		self.nonneg = True
		self.epsilon = epsilon

	def get_model_loss_fn(self):
		def _log_reg_loss(input, target):
			"""Input and target should be in the raw count space."""
			log_i = torch.log(input + self.epsilon)
			log_t = torch.log(target + self.epsilon)
			return torch.nn.MSELoss(reduction='mean')(log_i, log_t)
		return _log_reg_loss


class DeconvModelConvex(DeconvModel):
	"""Class for convex deconvolution models implemented using cvxpy.

	Attributes:
		dim_in (int): Input dimension, also number of groups.
		dim_out (int): Output dimension, also number of spots.
		model_name (str): Name of the deconvolution model.
		nonneg (bool): Whether to apply nonnegative contraint
			on spatial variables.
		deconv_configs (dict): Dictionary of deconvolution configurations.
		deconv_time (float): Time spent on deconvolution.
	"""
	def __init__(self, dim_in, dim_out, model_name) -> None:
		super().__init__()
		self.dim_in = dim_in # num_groups
		self.dim_out = dim_out # num_spots
		self.model_name = model_name
		self.nonneg = None # nonnegative constraint
		self.deconv_configs = None # deconvolution configs
		self.obj = None # objective function
		self.problem = None # convex problem to solve
		self.deconv_time = None # time spent on deconvolution

	def set_params(self, configs) -> bool:
		"""Set model parameters. Return True if warm start."""
		raise NotImplementedError

	def get_model_loss(self):
		"""Get regression reconstruction loss."""
		raise NotImplementedError

	def get_sp_loss(self):
		"""Get spatial loss (weighted by lambda_spatial_loss)."""
		raise NotImplementedError

	def get_props(self):
		"""Get the predicted celltype abundance from the trained deconvolution model."""
		raise NotImplementedError

	def deconv(self, x, y,
			   spatial_loss : SpatialLoss = None, lambda_spatial_loss = 0.0,
			   verbose = False, quiet = False, solver = None, **kwargs):
		"""Solve the regression-based spatial deconvolution problem using cvxpy.

		Loss_total = Loss_model + `lambda_spatial_loss` * Loss_spatial

		See https://www.cvxpy.org/tutorial/advanced/index.html#solve-method-options
		for solver options.

		Args:
			x (2D tensor): Bulk feature signiture matrix, num_feature x num_group.
			y (2D tensor): Spatial feature matrix, num_feature x num_spot.
			spatial_loss (SpatialLoss): The spatial smoothing loss.
				Spatial prior can be one of 'none', 'kl', 'sma', 'sar', 'car', 'icar'
				- KL : KL divergence of cell type proportion vectors of neiboring spots
				- SMA : Spatial moving average
				- SAR : Simultaneous auto-regressive model
				- CAR : Conditional auto-regressive model
				- ICAR : Intrinsic conditional auto-regressive model
				See model description for more details.
			lambda_spatial_loss (float): Specifies the strength of the spatial smoothing.
				Notice that setting `spatial_loss = None` will be more efficient than
				setting `lambda_spatial_loss = 0`, but the former won't benefit from
				warm start (i.e., speedup when only lambda_spatial_loss is changed).
			verbose (bool): If True, print solver output.
			quiet (bool): If True, no output printed, including final message.
			solver (str): The solver to use.
			kwargs: Additional keyword arguments specifying solver specific options.
		"""
  		# start timer
		t_start = timer()

		# set model parameters according to configs
		dim_in, dim_out = x.shape[1], y.shape[1]
		deconv_configs = {
			'x': x, 'y': y, 'dim_in': dim_in, 'dim_out': dim_out,
	  		'spatial_loss': spatial_loss, 'lambda_spatial_loss': lambda_spatial_loss,
			'verbose': verbose, 'quiet':quiet,
			'solver': solver, 'kwargs': kwargs
		}
		warm_start = self.set_params(deconv_configs)

		if not warm_start:
			# initialize optimization problem
			self.obj = cp.Minimize(self.get_model_loss() + self.get_sp_loss())
			self.problem = cp.Problem(self.obj)

		# pass solver arguments
		print_sol_out = verbose and (not quiet)
		self.problem.solve(solver = solver, verbose = print_sol_out,
						   warm_start=warm_start, **kwargs)

		# check problem status
		if not self.problem.status == cp.OPTIMAL:
			if self.problem.status == cp.UNBOUNDED:
				raise ValueError('Solver encountered numeric issues. '
								 'Try torch-based models or reduce number of genes.')
			warnings.warn(f"Problem status: {self.problem.status}.")

		# extract losses
		total_loss = self.obj.value
		model_loss = self.get_model_loss().value.item()
		if spatial_loss is None or lambda_spatial_loss == 0:
			sp_loss = 0
		else:
			sp_loss = self.get_sp_loss().value.item()

		t_end = timer()

		# save runtime
		self.deconv_time = t_end - t_start

		if not quiet: # print final message
			print(f'=== Time {self.deconv_time : .2f}s. '
				f'Loss: (total) {total_loss : .3f}, '
				f'(model) {model_loss : .3f}, '
				f'(spatial) {sp_loss : .3f}')


class LinearRegressionConvex(DeconvModelConvex):
	"""Linear regression implemented using cvxpy."""
	def __init__(self, bias = True, dim_in = 1, dim_out = 1):
		super().__init__(dim_in, dim_out, model_name = 'linear_cvx')
		self.nonneg = False
		self.use_bias = bias # whether to add the bias term
		if self.use_bias:
			self.bias = None

		# assume row-wise (group-wise) independency, aka each row is controlled
		# by an independent MVN distribution (can share the same cov).
		self.weights = None
		self.list_w_row = None

		# loss expressions
		self.model_loss_exp  = None # regression loss
		self.lambda_spatial_loss = None
		self.spatial_loss_exp = None # spatial_loss * lambda_spatial_loss

		self.warm_start = None # flag indicating whether to warm start

	def set_lambda_only(self, configs) -> bool:
		"""Check if the same configuration is used except lambda_spatial_loss.

		If True, will only update lambda_spatial_loss and enable warm start.
		"""
		if self.deconv_configs is not None:
			# check for shared configs and see whether to warm start
			l_is_same_item = [
				(configs['x'] == self.deconv_configs['x']).all(),
				(configs['y'] == self.deconv_configs['y']).all()
			] + [configs[key] == self.deconv_configs.get(key)
				 for key in (configs.keys() - \
					['x', 'y', 'lambda_spatial_loss', 'verbose', 'quiet'])]
			warm_start = all(l_is_same_item)

			if warm_start: # update only the lambda value to warm up the solver
				if configs['verbose'] and not configs['quiet']:
					print("Only the lambda_spatial_loss will be updated "
						  f"({self.deconv_configs['lambda_spatial_loss']:.3f} -> "
						  f"{configs['lambda_spatial_loss']:.3f}).")
				# update lambda
				self.lambda_spatial_loss.value = configs['lambda_spatial_loss']

				return True

		return False

	def set_params(self, configs) -> bool:
		# for sharing configs, update lambda only
		self.warm_start = self.set_lambda_only(configs)

		# update configurations
		self.deconv_configs = configs

		if self.warm_start: # exit if warm start
			return True

		# set model parameters to optimize
		self.dim_in = self.deconv_configs['dim_in']
		self.dim_out = self.deconv_configs['dim_out']
		self.list_w_row = [cp.Variable((1, self.dim_out), nonneg=self.nonneg)
						   for _ in range(self.dim_in)]
		self.weights = cp.vstack(self.list_w_row)

		# set reconstruction loss
		self.set_model_loss()

		# set spatial loss
		self.set_sp_loss()

		return False

	def set_model_loss(self):
		"""Set model reconstruction loss."""
		x, y = self.deconv_configs['x'], self.deconv_configs['y']

		if self.use_bias:
			self.bias = cp.Variable((1, self.dim_out))
			self.model_loss_exp = cp.sum_squares(x @ self.weights + self.bias - y) \
				/ (y.shape[0] * y.shape[1])
		else:
			self.model_loss_exp = cp.sum_squares(x @ self.weights - y) \
				/ (y.shape[0] * y.shape[1])

	def set_sp_loss(self):
		"""Set spatial loss."""
		sp_loss = self.deconv_configs['spatial_loss']
		# store as lambda as cp.Parameter for future warm start
		self.lambda_spatial_loss = cp.Parameter(nonneg=True)
		self.lambda_spatial_loss.value = self.deconv_configs['lambda_spatial_loss']
		self.spatial_loss_exp = 0

		# note that setting `spatial_loss = None` will be much faster than setting
		# `lambda_spatial_loss = 0`, but the former won't benefit from warm start
		if sp_loss is None: # exit if no spatial loss
			return None

		# extract inverse covariance matrix
		if not sp_loss.standardize_cov:
			raise ValueError("Spatial covariance must be standardized.")
		inv_cov = self.deconv_configs['spatial_loss'].inv_cov

		# use scipy sparse matrix to speed up computation
		if sp_loss.use_sparse:
			inv_cov = [psd_wrap(coo_matrix(
				(t.coalesce().values().numpy(), t.coalesce().indices().numpy()),
				shape=t.shape
    		)) for t in inv_cov]

		if len(inv_cov) == 1: # use the same covariance for all groups
			for w_row in self.list_w_row:
				self.spatial_loss_exp += \
					cp.quad_form(w_row.T, inv_cov[0]) * sp_loss.confidences[0]
		else: # use different covariances for different groups
			for i, w_row in enumerate(self.list_w_row):
				self.spatial_loss_exp += \
					cp.quad_form(w_row.T, inv_cov[i]) * sp_loss.confidences[i]

		self.spatial_loss_exp *= self.lambda_spatial_loss / (self.dim_in * self.dim_out)

	def get_model_loss(self):
		"""Get model regression reconstruction loss."""
		return self.model_loss_exp

	def get_sp_loss(self):
		"""Get spatial loss (weighted by lambda_spatial_loss)."""
		return self.spatial_loss_exp

	def get_props(self):
		"""Get the predicted celltype abundance from the trained deconvolution model."""
		# remove negative abundances
		props = np.clip(self.weights.value, 0, None).T # num_out x num_in
		# normalize
		props = props / (props.sum(1, keepdims=True) + 1e-8)
		return props


class NNLSConvex(LinearRegressionConvex):
	"""Non-negative least square implemented using cvxpy."""
	def __init__(self, bias = True, dim_in = 1, dim_out = 1):
		super().__init__(bias, dim_in, dim_out)
		self.model_name = 'nnls_cvx'
		self.nonneg = True


class DWLSConvex(LinearRegressionConvex):
	"""Damped weighted least square implemented using cvxpy.

	Here the error term for each feature is scaled by its observed value, not the predicted
	value as described in https://www.nature.com/articles/s41467-019-10802-z.

	Attributes:
		max_weights (float): The upper limit of the scaling weights for regression loss.
	"""
	def __init__(self, bias = True, dim_in = 1, dim_out = 1, nonneg = True, max_weights = 8) -> None:
		super().__init__(bias, dim_in, dim_out)
		self.model_name = 'dwls_cvx'
		self.nonneg = nonneg
		self.max_weights = max_weights

	def set_model_loss(self):
		"""Set DWLS regression loss."""
		x, y = self.deconv_configs['x'], self.deconv_configs['y']
		# calculate weights for each gene at each spot
		model_loss_scales = (y.pow(-2)).clamp(max = self.max_weights ** 2) # num_feature x num_spot
		# set the minimum weight to 1
		min_weights, _ = torch.min(model_loss_scales, dim=0, keepdim=True)
		model_loss_scales = (model_loss_scales / min_weights).clamp(max = self.max_weights)

		if self.use_bias:
			self.bias = cp.Variable((1, self.dim_out))
			model_loss_exp = cp.square(x @ self.weights + self.bias - y)
		else:
			model_loss_exp = cp.square(x @ self.weights - y)

		self.model_loss_exp = cp.sum(cp.multiply(model_loss_exp, model_loss_scales)) \
				/ (y.shape[0] * y.shape[1])


class NuSVRConvex(LinearRegressionConvex):
	"""Nu-Support vector regression implemented using cvxpy.

	Attributes:
		C (float): SVR regularization parameter.
		nu (float): Nu-SVR parameter.
		epsilon (float): Epsilon in the epsilon-SVR model, specifying the the epsilon-tube within which
			no penalty is associated in the training loss function with points predicted within a distance
			epsilon from the actual value.
		loss_mode (str): Specifies the regression loss function to use. Either 'l2' or 'l1'.
	"""
	def __init__(self, bias = True, dim_in = 1, dim_out = 1,
			  	 C = 1.0, nu = 0.1, nonneg = False, loss_mode = 'l2'):
		"""Initialize a Nu-SVR deconvolution model.

		Args:
			C (float): SVR regularization parameter.
			nu (float): Nu-SVR parameter.
			nonneg (bool): Whether to require nonnegative regression coefficients.
			loss_mode (str): Specifies the regression loss function to use. Either 'l2' or 'l1'.
		"""
		super().__init__(bias, dim_in, dim_out)

		# store model configs
		self.model_name = 'svr_cvx'
		self.nonneg = nonneg

		# check regression loss fn, 'l2' or 'l1', default 'l2' for faster convergence
		if loss_mode not in ['l2', 'l1']:
			raise ValueError("loss_mode must be either 'l2' or 'l1'.")
		self.loss_mode = loss_mode

		self.C = C
		self.nu = cp.Parameter(nonneg=True)
		self.nu.value = nu
		self.epsilon = None

	def set_model_loss(self):
		"""Set Nu-SVR regression loss."""
		x, y = self.deconv_configs['x'], self.deconv_configs['y']
		self.epsilon = cp.Variable(nonneg=True)

		# calculate regression loss
		if self.use_bias:
			self.bias = cp.Variable((1, self.dim_out))
			res = cp.pos(cp.abs(x @ self.weights + self.bias - y) - self.epsilon)
		else:
			res = cp.pos(cp.abs(x @ self.weights - y) - self.epsilon)

		# use either L2 or L1 loss
		if self.loss_mode == 'l2':
			model_loss = cp.sum_squares(res) / (y.shape[0] * y.shape[1])
		else:
			model_loss = cp.sum(res) / (y.shape[0] * y.shape[1])

		# add regularization loss for epsilon and weights
		self.model_loss_exp = self.C * (self.nu * self.epsilon + model_loss) + \
			0.5 * cp.sum_squares(self.weights) / (self.dim_in * self.dim_out)

	def set_params(self, configs) -> bool:
		# for sharing configs, update lambda only
		self.warm_start = self.set_lambda_only(configs)

		# update configurations
		self.deconv_configs = configs

		if self.warm_start: # exit if warm start
			return True

		# set model parameters to optimize
		self.dim_in = self.deconv_configs['dim_in']
		self.dim_out = self.deconv_configs['dim_out']
		self.list_w_row = [cp.Variable((1, self.dim_out), nonneg=self.nonneg)
						   for _ in range(self.dim_in)]
		self.weights = cp.vstack(self.list_w_row)

		# set reconstruction loss
		self.set_model_loss()

		# set spatial loss
		self.set_sp_loss()

		return False



"""
Simulate spatial patterns from 2D Gaussian process.
Some functions adapted from the cell2location paper:
	Kleshchevnikov, Vitalii, et al. "Cell2location maps fine-grained cell types in spatial transcriptomics."
	Nature biotechnology 40.5 (2022): 661-671.
	https://www.nature.com/articles/s41587-021-01139-4#
"""

import itertools
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.stats import norm
from tqdm import tqdm


def kernel(X1, X2, band_width=1.0, eta=1.0):
	'''
	Isotropic squared exponential kernel. Computes
	a covariance matrix from points in X1 and X2.

	Args:
		X1: Array of m points (m x d).
		X2: Array of n points (n x d).
		band_width: Bandwidth of the Gaussian kernel.

	Returns:
		Covariance matrix (m x n).
	'''
	sqdist = np.sum(X1**2, 1).reshape(-1, 1) + np.sum(X2**2, 1) - 2 * np.dot(X1, X2.T)
	return eta**2 * np.exp(-0.5 / band_width**2 * sqdist)

def generate_grid(grid_size=[50, 50]):
	'''
	Generating a grid containing n_spots
	for each spot, return the x1 and x2 pos
	'''
	n1, n2 = grid_size
	x1 = np.linspace(0, 100, n1)[:,None]
	x2 = np.linspace(0, 100, n2)[:,None]

	# make cartesian grid out of each dimension x1 and x2
	grid = np.array(list(itertools.product(x1, x2)))
	return (grid, x1, x2)

def random_gp(coords, x1 = 1, x2 = 1, #coordinates
			  n_patterns = 3, # number of zones to simulate
			  eta_true = 5, #variance, defines overlapping
			  l1_true = [8, 10, 15], #bw parameter
			  l2_true = [8, 10, 15]
			 ):
	'''
	Generating random 2D gaussian based on an isotropic squared exponential kernel.

	Args:
		coords (2D array): spatial coordinates for each spot (n_spots x 2).
		x1 (1D array): x-coordinates of the grid (x_size x 1).
		x2 (1D array): y-coordinates of the grid (y_size x 1).
	'''
	assert len(l1_true) == len(l2_true) == n_patterns

	# generate spatial covariance matrix
	#cov1, cov2 = kernel(x1, x1, l1_true), kernel(x2, x2, l2_true)
	sp_cov_list = [np.kron(kernel(x1, x1, band_width=l1_true[i], eta=eta_true),
						   kernel(x2, x2, band_width=l2_true[i], eta=eta_true))
				   for i in range(n_patterns)]

	# sample abundances from GP
	gaus_true = np.stack(
		[np.random.multivariate_normal(np.zeros(coords.shape[0]), # mean
									   2 * sp_cov_list[i])  # cov
		 for i in range(n_patterns)]
	).T

	# softmax transformation
	props_true = (np.exp(gaus_true).T / np.exp(gaus_true).sum(axis=1)).T
	return props_true


def plot_spatial(values, grid_size=[50,50], nrows=5, names=['cell type'],
				 vmin=0, vmax=1, cmap="magma"):
	'''
	Plot the input spatial value in 2D
	'''
	if len(values.shape) == 1:
		values = np.array(values).reshape(-1,1)
	n_cell_types = values.shape[1]
	n1, n2 = grid_size
	ncols = np.ceil((n_cell_types+1) / nrows).astype('int')
	for ct in range(n_cell_types):
		plt.subplot(nrows, ncols, ct+1)
		plt.imshow(values[:,ct].reshape(n1,n2).T,
				   cmap=plt.cm.get_cmap(cmap),
				   vmin=vmin, vmax=vmax
				  )
		plt.colorbar()
		if len(names) > 1:
			plt.title(names[ct])
		else:
			plt.title(f'{names[0]} {ct+1}')

	plt.subplot(nrows, ncols, n_cell_types+1)
	plt.imshow(values.sum(axis=1).reshape(n1,n2).T,
			   cmap=plt.cm.get_cmap('Greys'))
	plt.colorbar()
	plt.title('total')


def sample_gp(locations, x1, x2,
			  n_zones, # number of zones
			  band_width,
			  eta = 1.5):
	'''
	Sample abundances with 2D GP
	'''

	assert len(band_width) == n_zones
	if n_zones == 0:
		return np.empty([locations.shape[0],0])

	sparse_abundances = random_gp(coords=locations, x1=x1, x2=x2, n_patterns=n_zones,
								  eta_true=eta,l1_true=band_width,l2_true=band_width)

	sparse_abundances = sparse_abundances / sparse_abundances.max(0)
	sparse_abundances[sparse_abundances < 0.1] = 0

	return sparse_abundances


def conc_cir(x1,x2, mu_r=0, sd=0.1):
	'''
	Sample concentric circlar patterns
	'''
	X1 = 2*((x1-np.mean(x1))/np.max(x1))**2
	X2 = 2*((x2-np.mean(x2))/np.max(x2))**2

	d = np.stack([X1 for i in range(len(x2))]) \
	+ np.stack([X2 for i in range(len(x1))], axis = 1)

	return norm(mu_r, sd).pdf(np.sqrt(d)).reshape(-1,1)

def sample_circ(locations, x1, x2,
				n_circ, # number of circular (regional) zones
				r_circ, # list of means of radius
				sd_circ, # list of sds of radius
				band_width,
				eta = 0.5):
	'''
	Sample abundances with GP based on a prior concentric circlar pattern
	'''
	## regional regions
	assert len(r_circ) == n_circ
	assert len(sd_circ) == n_circ
	assert len(band_width) == n_circ

	# looping through all circular zones
	cir_abundances = np.stack([conc_cir(x1, x2, r_circ[i], sd_circ[i])
                               for i in range(n_circ)], axis = 1)


	gp_abundances = random_gp(coords=locations, x1=x1, x2=x2, n_patterns=n_circ,
													eta_true=eta,
													l1_true=band_width,
													l2_true=band_width)

	sparse_abundances = cir_abundances.reshape(cir_abundances.shape[0:2]) * gp_abundances

	sparse_abundances = sparse_abundances / sparse_abundances.max(0)
	sparse_abundances[sparse_abundances < 0.1] = 0

	return sparse_abundances

def sample_hist(locations, x1, x2,
				n_hist, # number of histology clusters (zones)
				hist_abundances, # membership matrix
				band_width, eta # GP parameters
				):
	'''
	Sample abundances with GP based on a prior histological pattern
	'''
	gp_abundances = random_gp(coords=locations, x1=x1, x2=x2, n_patterns=n_hist,
													eta_true=eta,
													l1_true=band_width,
													l2_true=band_width)

	sparse_abundances = hist_abundances * gp_abundances

	sparse_abundances = sparse_abundances / sparse_abundances.max(0)
	sparse_abundances[sparse_abundances < 0.1] = 0

	return sparse_abundances

def get_abundances(locations, x1, x2, # location coordinates for 1 experiment
				   n_regional_zones, n_ubiquitous_zones, # zone information
				   band_width, # GP parameters
				   n_experiments, n_locations,
				   plot_abundances = True,
				   eta_rg = 1.5, # GP parameters
				   eta_ub = 0.3
				   ):
	'''
	Get the abundances matrix as presented in the cell2location notebook

	Args:
		locations, x1, x2: # location coordinates for 1 experiment
		n_regional_zones: number of regional zones
		n_ubiquitous_zones: number of ubiquitous zones
		band_width, eta_rg, eta_ub: parameters used to calculate the 2D-gaussian kernel
		n_experiments: number of experiments required
		n_locations: number of locations[n_row, n_col] per experiment
		plot_abundances: whether to plot the abundance patterns of each zone in the first experiment

	Returns:
		abundances_df: (np.prod(n_locations) * n_exp) x (n_regional_zones + n_ubiquitous_zones)
	'''
	abundances_df = pd.DataFrame()
	for exp in tqdm(range(n_experiments)):
		# abundance df (n_locations, n_zones)
		sparse_abundances = sample_gp(locations, x1, x2,
					n_regional_zones, # number of circular (regional) zones
					band_width[0:n_regional_zones],
					eta = eta_rg)

		uniform_abundances = sample_gp(locations, x1, x2, n_ubiquitous_zones,
									   band_width[n_regional_zones:(n_regional_zones+n_ubiquitous_zones)], eta = eta_ub)

		abundances_df_1 = np.concatenate([sparse_abundances, uniform_abundances], axis=1)
		abundances_df_1 = pd.DataFrame(abundances_df_1,
									   index=[f'exper{exp}_location_{i}' for i in range(abundances_df_1.shape[0])],
									   columns=[f'tissue_zone_{i}' for i in range(n_regional_zones)] \
									   + [f'uniform_{i}' for i in range(n_ubiquitous_zones)])

		abundances_df = pd.concat((abundances_df, abundances_df_1), axis=0)

		if (exp == 0) & (plot_abundances) :
			plt.figure(figsize=(3*5+5,3*5+5))
			plot_spatial(abundances_df_1.values, grid_size=n_locations, nrows=5, names=abundances_df.columns)
			plt.show()
	return abundances_df

def get_abundances_circ(locations, x1, x2, # location coordinates for 1 experiment
						n_regional_zones, n_ubiquitous_zones, # zone information
						r_circ, sd_circ, band_width, # GP parameters
						n_experiments, n_locations,
						plot_abundances = True,
						eta = 0.3 # GP parameters
						):
	'''
	Get the abundances matrix using donut patterns

	Args:
		locations, x1, x2: # location coordinates for 1 experiment
		n_regional_zones: number of regional zones
		n_ubiquitous_zones: number of ubiquitous zones
		r_circ, sd_circ: the mean and std of a gaussian distribution used to sample circular ring patterns
		band_width, eta: parameters used to calculate the 2D-gaussian kernel
		n_experiments: number of experiments required
		n_locations: number of locations[n_row, n_col] per experiment
		plot_abundances: whether to plot the abundance patterns of each zone in the first experiment

	Returns:
		abundances_df: (np.prod(n_locations) * n_exp) x (n_regional_zones + n_ubiquitous_zones)
	'''
	abundances_df = pd.DataFrame()
	for exp in tqdm(range(n_experiments)):
		# abundance df (n_locations, n_zones)
		sparse_abundances = sample_circ(locations, x1, x2,
					n_regional_zones, # number of circular (regional) zones
					r_circ, # list of means of radius
					sd_circ, # list of sds of radius
					band_width[0:n_regional_zones],
					eta = eta)

		uniform_abundances = sample_gp(locations, x1, x2, n_ubiquitous_zones,
									   band_width[n_regional_zones:(n_regional_zones+n_ubiquitous_zones)], eta = eta)

		abundances_df_1 = np.concatenate([sparse_abundances, uniform_abundances], axis=1)
		abundances_df_1 = pd.DataFrame(abundances_df_1,
									   index=[f'exper{exp}_location_{i}' for i in range(abundances_df_1.shape[0])],
									   columns=[f'tissue_zone_{i}' for i in range(n_regional_zones)] \
									   + [f'uniform_{i}' for i in range(n_ubiquitous_zones)])

		abundances_df = pd.concat((abundances_df, abundances_df_1), axis=0)

		if (exp == 0) & (plot_abundances) :
			plt.figure(figsize=(3*5+5,3*5+5))
			plot_spatial(abundances_df_1.values, grid_size=n_locations, nrows=5, names=abundances_df.columns)
			plt.show()
	return abundances_df

def get_abundances_hist(locations, x1, x2, # location coordinates for 1 experiment
						n_regional_zones, n_ubiquitous_zones, # zone information
						hist_abundances,
						band_width, # GP parameters
						n_experiments, n_locations,
						plot_abundances = True,
						eta = 0.3 # GP parameters
						):
	'''
	Get the abundances matrix using histological patterns

	Args:
		locations, x1, x2: # location coordinates for 1 experiment
		n_regional_zones: number of regional zones
		n_ubiquitous_zones: number of ubiquitous zones
		hist_abundances: membership matrix from fuzzy-c-means clustering,
  			np.prod(n_locations) x n_regional_zones
		band_width, eta: parameters used to calculate the 2D-gaussian kernel
		n_experiments: number of experiments required
		n_locations: number of locations[n_row, n_col] per experiment
		plot_abundances: whether to plot the abundance patterns of each zone in the first experiment

	Returns:
		abundances_df: (np.prod(n_locations) * n_exp) x (n_regional_zones + n_ubiquitous_zones)
	'''
	abundances_df = pd.DataFrame()
	for exp in tqdm(range(n_experiments)):
		# abundance df (n_locations, n_zones)
		sparse_abundances = sample_hist(locations, x1, x2,
										n_regional_zones, # number of histology clusters (zones)
										hist_abundances, # membership matrix
										band_width[0:n_regional_zones],eta # GP parameters
									   )

		uniform_abundances = sample_gp(locations, x1, x2, n_ubiquitous_zones,
									   band_width[n_regional_zones:(n_regional_zones+n_ubiquitous_zones)], eta = eta)

		abundances_df_1 = np.concatenate([sparse_abundances, uniform_abundances], axis=1)
		abundances_df_1 = pd.DataFrame(abundances_df_1,
									   index=[f'exper{exp}_location_{i}' for i in range(abundances_df_1.shape[0])],
									   columns=[f'tissue_zone_{i}' for i in range(n_regional_zones)] \
									   + [f'uniform_{i}' for i in range(n_ubiquitous_zones)])

		abundances_df = pd.concat((abundances_df, abundances_df_1), axis=0)

		if (exp == 0) & (plot_abundances) :
			plt.figure(figsize=(3*5+5,3*5+5))
			plot_spatial(abundances_df_1.values, grid_size=n_locations, nrows=5, names=abundances_df.columns)
			plt.show()
	return abundances_df



"""
Generate ST count matrix based on the assigned pattern.
Some functions adapted from the cell2location paper:
	Kleshchevnikov, Vitalii, et al. "Cell2location maps fine-grained cell types in spatial transcriptomics."
	Nature biotechnology 40.5 (2022): 661-671.
	https://www.nature.com/articles/s41587-021-01139-4#
"""
from re import sub
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix,lil_matrix
import anndata
from tqdm import tqdm
from smoother.weights import coordinate_to_weights_knn_dense


def sample_cell_indices(generation_snrna, annot_label, cell_count_df, cell_capture_eff_df):
	'''
	Sample cell indices function adapted from cell2location paper
	(https://github.com/vitkl/cell2location_paper/blob/master/notebooks/benchmarking/synthetic_data_construction_improved_real_mg.ipynb)

	Args:
		generation_snrna: reference single-cell adata.
		cell_count_df, cell_capture_eff_df: Array of n_spots x n_cell_types

	Returns:
		cell index matrix (n_spots x n_cells).
	'''
	# num_spots x num_cells_in_train_set
	locations2cells = np.zeros((cell_count_df.shape[0], generation_snrna.shape[0]))

	for i, l in enumerate(cell_count_df.index): # locations
		for ct in cell_count_df.columns: # celltypes
			cell_ind_all = generation_snrna.obs['cell_ind']
			cell_ind_all = cell_ind_all[generation_snrna.obs[annot_label] == ct]
			# assign single cells to each location
			cell_ind = np.random.choice(cell_ind_all, int(cell_count_df.loc[l, ct]), replace=False)

			# scale by capture efficiency
			locations2cells[i,cell_ind] = cell_capture_eff_df.loc[l, ct]

	return csr_matrix(locations2cells)

def diag_block_mat_boolindex(L):
	'''
	Helper function to generate diagonal matrix across n_experiments
	'''
	shp = L[0].shape
	mask = np.kron(np.eye(len(L)), np.ones(shp))==1
	out = np.zeros(np.asarray(shp)*len(L),dtype=float)
	out[mask] = np.concatenate(L).ravel()
	return csr_matrix(out)

def smooth_by_neighbors(shared_prop, locations2cells_matrix, coords, n_experiments):
	'''
	Smoothing the spot x cell-index matrix by neighoring spots

	Args:
		shared_prop: proportion of cells that come from neighboring spots.
		locations2cells_matrix: binary spot x cell-index matrix,
			((n_spots * n_experiments) x n_cells).
		coords: n_spots x 2 (x and y coordinates) grids, the same for all experiments.
		n_experiments: number of experiments.

	Returns:
		smoothed cell index matrix (n_spots x n_cells).
	'''
	# compute adjacency matrix
	swm = coordinate_to_weights_knn_dense(coords, k=4, row_scale=True) * shared_prop
	swm = swm.fill_diagonal_(1 - shared_prop)

	# compute smoothed cell index matrix
	swm_all_exp = diag_block_mat_boolindex([swm for i in range(n_experiments)])
	locations2cells_smoothed = swm_all_exp.dot(locations2cells_matrix)
	return csr_matrix(locations2cells_smoothed)

def create_spatial_adata(adata, locations2cells_matrix,
						 cell_counts, cell_abundances, capture_eff,
						 locations, cell_types2zones_df):
	'''
	Construct count matrix using locations x cell_indice matrix and reference anndata

	Args:
		adata: refrence adata where the raw counts are stored in adata.X
		locations2cells_matrix: cell index matrix (n_spots x n_cells)
		locations: (n_experiments * n_spots) x 2 (x and y coordinates) grids
		cell_abundances,cell_count_df, cell_capture_eff_df: Array of n_spots x n_cell_types
		cell_types2zones_df: n_cell_types x n_zones matrix

	Returns:
		synthetic anndata object
	'''
	print("Computing synthetic counts" )
	synthetic_counts = locations2cells_matrix.dot(adata.X)

	print("Constructing adata object......")
	# Create adata object
	synth_adata = anndata.AnnData(synthetic_counts, dtype = np.dtype('float'))
	synth_adata.obs_names = cell_counts.index
	synth_adata.var_names = adata.var_names
	synth_adata.obs[[f'cell_count_{ct}' for ct in cell_counts.columns]] = cell_counts
	synth_adata.obs[[f'cell_abundances_{ct}' for ct in cell_abundances.columns]] = cell_abundances
	synth_adata.obs[[f'cell_capture_eff_{ct}' for ct in capture_eff.columns]] = capture_eff
	synth_adata.obsm['X_spatial'] = locations
	synth_adata.uns['design'] = {'cell_types2zones': cell_types2zones_df}
	synth_adata.obs['sample'] = [sub('_location.+$','',i) for i in synth_adata.obs_names]
	return synth_adata

def cal_counts_ct(synth_adata,
				  orig_adata,
				  annot_label,
				  locations2cells_matrix):
	'''
	Compute synthetic counts per cell type

	Args:
		synth_adata: synthetic adata
		orig_adata: reference adata
		annot_label: key to access celltype information
		locations2cells_matrix: cell index matrix (n_spots x n_cells)

	Returns:
		updated synth_adata with UMI per cell_type stored in obs

	'''
	cell_types = np.unique(orig_adata.obs[annot_label])
	for _, ct in tqdm(enumerate(cell_types), total = len(cell_types)):
		select = np.where(orig_adata.obs[annot_label]==ct)[0]
		locations2cells_ct = lil_matrix(np.zeros(locations2cells_matrix.shape))
		locations2cells_ct[:,select] = locations2cells_matrix[:,select]
		synthetic_counts_ct = locations2cells_ct.tocsr().dot(orig_adata.X)
		synth_adata.obs[f'UMI_count_{ct}'] = np.array(synthetic_counts_ct.sum(1)).flatten()

	return synth_adata

def add_gamma_noises(synth_adata,
					 shape = 0.4586,
					 scale=1/0.6992
					):
	'''
	Fuction for adding noises adapted from cell2location paper
	(https://github.com/vitkl/cell2location_paper/blob/master/notebooks/benchmarking/synthetic_data_construction_improved_real_mg.ipynb)

	Args:
		synth_adata: synthetic adata
		shape: shape of gamma distribution
		scale: scale of gamma distribution

	Returns:
		synthetic adata
	'''
	# Sample detection rates
	gene_level = np.random.gamma(shape,scale=scale,
								 size=(1, synth_adata.shape[1])) / 2
	synth_adata.var['gene_level'] = gene_level.flatten()
	synth_adata.X = synth_adata.X.toarray() * gene_level

	# Sample poisson integers
	synth_adata.layers['expression_levels'] = synth_adata.X
	synth_adata.X = np.random.poisson(synth_adata.X)
	synth_adata.X = csr_matrix(synth_adata.X)

	return synth_adata

def grouped_obs_mean(adata, group_key, layer=None, gene_symbols=None):
	'''
	Group referenceobs by celltypes and output group mean for deconvolution

	Args:
		adata: reference anndata
		group_key: key to access celltype information
		layer: anndata layer that stores the feature matrix to calculate group mean
		gene_symbols: genes to keep; keep all genes when "none"

	Returns:
		gene x cell_type count matrix
	'''
	if layer is not None: # if layer is provided, use it
		get_x = lambda x: x.layers[layer]
	else: # load data stored in adata.X
		get_x = lambda x: x.X
	if gene_symbols is not None:
		feature_idx = gene_symbols
	else:
		feature_idx = adata.var_names

	grouped = adata.obs.groupby(group_key)
	out = pd.DataFrame(
		np.zeros((len(feature_idx), len(grouped)), dtype=np.float64),
		columns=list(grouped.groups.keys())
	)

	for group, idx in grouped.indices.items():
		ref_counts = get_x(adata[idx, feature_idx])
		out[group] = np.ravel(ref_counts.mean(axis=0, dtype=np.float64))

	out.index = feature_idx
	return out



import logging
from typing import Iterable, List, Literal, Optional, Union, Sequence
from copy import deepcopy
import warnings

import numpy as np
import torch
from torch.distributions import kl_divergence as kl
from torch.distributions import Categorical, Normal

from scvi import REGISTRY_KEYS
from scvi._types import Tunable
from scvi.module.base import LossOutput
from scvi.module import SCANVAE
from scvi.model import SCVI, SCANVI
from scvi.train import TrainRunner, TrainingPlan
from scvi.data._constants import _SETUP_ARGS_KEY

from anndata import AnnData
from scvi.dataloaders import DeviceBackedDataSplitter
from scvi.utils._docstrings import devices_dsp

from smoother import SpatialLoss
from ._utils import broadcast_labels, get_max_epochs_heuristic, set_params_online_update


logger = logging.getLogger(__name__)


class SPANVAE(SCANVAE):
	"""Add spatial loss to the latent representation in the SCANVAE model.

	Parameters:
		spatial_loss: Spatial loss to apply on the latent representation. If None, no spatial loss.
		lambda_spatial_loss: Weight of the spatial loss.
		sp_loss_on: Whether to apply the spatial loss on the sampled latent representation ('z') or
			on the encoded mean of the latent distribution ('mean_z').
	"""

	def __init__(
		self,
		n_input: int,
		spatial_loss: Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1,
		sp_loss_on: Literal['z', 'mean_z'] = 'mean_z',
		n_batch: int = 0,
		n_labels: int = 0,
		n_hidden: Tunable[int] = 128,
		n_latent: Tunable[int] = 10,
		n_layers: Tunable[int] = 1,
		n_continuous_cov: int = 0,
		n_cats_per_cov: Optional[Iterable[int]] = None,
		dropout_rate: Tunable[float] = 0.1,
		dispersion: Tunable[
			Literal["gene", "gene-batch", "gene-label", "gene-cell"]
		] = "gene",
		log_variational: Tunable[bool] = True,
		gene_likelihood: Tunable[Literal["zinb", "nb"]] = "zinb",
		y_prior=None,
		labels_groups: Sequence[int] = None,
		use_labels_groups: bool = False,
		linear_classifier: bool = False,
		classifier_parameters: Optional[dict] = None,
		use_batch_norm: Tunable[Literal["encoder", "decoder", "none", "both"]] = "both",
		use_layer_norm: Tunable[Literal["encoder", "decoder", "none", "both"]] = "none",
		**vae_kwargs,

	):
		# initialize the SCANVAE model
		super().__init__(
			n_input=n_input,
			n_batch=n_batch,
			n_labels=n_labels,
			n_hidden=n_hidden,
			n_latent=n_latent,
			n_layers=n_layers,
			n_continuous_cov=n_continuous_cov,
			n_cats_per_cov=n_cats_per_cov,
			dropout_rate=dropout_rate,
			dispersion=dispersion,
			log_variational=log_variational,
			gene_likelihood=gene_likelihood,
			y_prior=y_prior,
			labels_groups=labels_groups,
			use_labels_groups=use_labels_groups,
			linear_classifier=linear_classifier,
			classifier_parameters=classifier_parameters,
			use_batch_norm=use_batch_norm,
			use_layer_norm=use_layer_norm,
			**vae_kwargs,
		)

		# spatial loss
		self.spatial_loss = spatial_loss
		self.l_sp_loss = lambda_spatial_loss
		self.sp_loss_on = sp_loss_on

	def loss(
		self,
		tensors,
		inference_outputs,
		generative_ouputs,
		feed_labels=False,
		kl_weight=1,
		labelled_tensors=None,
		classification_ratio=None,
	):
		"""Compute the loss."""
		# from the original SCANVAE loss
		# unpack hidden and latent representation
		px = generative_ouputs["px"]
		qz1 = inference_outputs["qz"]
		z1 = inference_outputs["z"]
		x = tensors[REGISTRY_KEYS.X_KEY]
		batch_index = tensors[REGISTRY_KEYS.BATCH_KEY]

		if feed_labels:
			y = tensors[REGISTRY_KEYS.LABELS_KEY]
		else:
			y = None
		is_labelled = False if y is None else True

		# Enumerate choices of label
		ys, z1s = broadcast_labels(y, z1, n_broadcast=self.n_labels)
		qz2, z2 = self.encoder_z2_z1(z1s, ys)
		pz1_m, pz1_v = self.decoder_z1_z2(z2, ys)
		reconst_loss = -px.log_prob(x).sum(-1)

		# KL Divergence
		mean = torch.zeros_like(qz2.loc)
		scale = torch.ones_like(qz2.scale)

		kl_divergence_z2 = kl(qz2, Normal(mean, scale)).sum(dim=1)
		loss_z1_unweight = -Normal(pz1_m, torch.sqrt(pz1_v)).log_prob(z1s).sum(dim=-1)
		loss_z1_weight = qz1.log_prob(z1).sum(dim=-1)
		if not self.use_observed_lib_size:
			ql = inference_outputs["ql"]
			(
				local_library_log_means,
				local_library_log_vars,
			) = self._compute_local_library_params(batch_index)

			kl_divergence_l = kl(
				ql,
				Normal(local_library_log_means, torch.sqrt(local_library_log_vars)),
			).sum(dim=1)
		else:
			kl_divergence_l = 0.0

		# add the spatial loss
		if self.spatial_loss is not None and self.l_sp_loss > 0.0:
			if self.sp_loss_on == 'z':
				sp_loss = self.spatial_loss(z1.T, normalize = False)
			elif self.sp_loss_on == 'mean_z':
				sp_loss = self.spatial_loss(qz1.loc.T, normalize = False)
			else:
				raise NotImplementedError('Currently the spatial loss can only be applied on z or mean_z.')

			sp_loss *= self.l_sp_loss / self.n_latent
		else:
			sp_loss = torch.tensor(0.0)

		if is_labelled:
			loss = reconst_loss + loss_z1_weight + loss_z1_unweight + sp_loss
			kl_locals = {
				"kl_divergence_z2": kl_divergence_z2,
				"kl_divergence_l": kl_divergence_l,
			}
			if labelled_tensors is not None:
				ce_loss, true_labels, logits = self.classification_loss(
					labelled_tensors
				)
				loss += ce_loss * classification_ratio
				return LossOutput(
					loss=loss,
					reconstruction_loss=reconst_loss,
					kl_local=kl_locals,
					kl_global=sp_loss,
					classification_loss=ce_loss,
					true_labels=true_labels,
					logits=logits,
					extra_metrics={
						"n_labelled_tensors": labelled_tensors[
							REGISTRY_KEYS.X_KEY
						].shape[0],
					},
				)
			return LossOutput(
				loss=loss,
				reconstruction_loss=reconst_loss,
				kl_local=kl_locals,
				kl_global=sp_loss,
			)

		probs = self.classifier(z1)
		reconst_loss += loss_z1_weight + (
			(loss_z1_unweight).view(self.n_labels, -1).t() * probs
		).sum(dim=1)

		kl_divergence = (kl_divergence_z2.view(self.n_labels, -1).t() * probs).sum(
			dim=1
		)
		kl_divergence += kl(
			Categorical(probs=probs),
			Categorical(probs=self.y_prior.repeat(probs.size(0), 1)),
		)
		kl_divergence += kl_divergence_l

		loss = torch.mean(reconst_loss + kl_divergence * kl_weight) + sp_loss

		if labelled_tensors is not None:
			ce_loss, true_labels, logits = self.classification_loss(labelled_tensors)

			loss += ce_loss * classification_ratio
			return LossOutput(
				loss=loss,
				reconstruction_loss=reconst_loss,
				kl_local=kl_divergence,
				kl_global=sp_loss,
				classification_loss=ce_loss,
				true_labels=true_labels,
				logits=logits,
			)
		return LossOutput(
			loss=loss, reconstruction_loss=reconst_loss,
			kl_local=kl_divergence,
			kl_global=sp_loss,
		)

	@classmethod
	def from_vae(
		cls,
		vae_module: SCANVAE,
		spatial_loss: Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1,
		sp_loss_on: Literal['z', 'mean_z'] = 'mean_z'
	):

		# copy the VAE module
		spanvae = deepcopy(vae_module)

		# switch to the spatial module
		spanvae.__class__ = cls
		spanvae.spatial_loss = spatial_loss
		spanvae.l_sp_loss = lambda_spatial_loss
		spanvae.sp_loss_on = sp_loss_on

		return spanvae



class SpatialANVI(SCANVI):
	"""Spatially-aware ANnotation Variational Inference model.
	"""

	_data_splitter_cls = DeviceBackedDataSplitter
	_module_cls = SPANVAE
	_training_plan_cls = TrainingPlan # UnsupervisedTrainingPlan
	_train_runner_cls = TrainRunner

	def __init__(
		self,
		st_adata: AnnData,
		spatial_loss: Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1,
		sp_loss_on: Literal['z', 'mean_z'] = 'mean_z',
		n_hidden: int = 128,
		n_latent: int = 10,
		n_layers: int = 1,
		dropout_rate: float = 0.1,
		dispersion: Literal["gene", "gene-batch", "gene-label", "gene-cell"] = "gene",
		gene_likelihood: Literal["zinb", "nb", "poisson"] = "nb",
		latent_distribution: Literal["normal", "ln"] = "normal",
		**model_kwargs,
	):
		super().__init__(
			st_adata,
			n_hidden=n_hidden,
			n_latent=n_latent,
			n_layers=n_layers,
			dropout_rate=dropout_rate,
			dispersion=dispersion,
			gene_likelihood=gene_likelihood,
			latent_distribution=latent_distribution,
			**model_kwargs
		)

		self.module = self._module_cls.from_vae(
			self.module,
			spatial_loss=spatial_loss,
			lambda_spatial_loss=lambda_spatial_loss,
			sp_loss_on=sp_loss_on
		)

		self._model_summary_string = self._model_summary_string.replace('ScanVI', 'SpatialANVI')
		self.init_params_ = self._get_init_params(locals())
		self.dr_logs = {'elapsed_time': None, 'total_loss': [],
						'recon_loss': [], 'spatial_loss': []}



	@devices_dsp.dedent
	def train(
		self,
		max_epochs: Optional[int] = None,
		accelerator: str = "auto",
		devices: Union[int, List[int], str] = "auto",
		plan_kwargs: Optional[dict] = None,
		**trainer_kwargs,
	):
		"""Unsupervised training of the spatial model without minibatches."""
		if max_epochs is None:
			max_epochs = get_max_epochs_heuristic(self.adata.n_obs)

			if self.was_pretrained:
				max_epochs = int(np.min([10, np.max([2, round(max_epochs / 3.0)])]))

		logger.info(f"Training for {max_epochs} epochs.")

		plan_kwargs = {} if plan_kwargs is None else plan_kwargs

		data_splitter = self._data_splitter_cls(
			adata_manager=self.adata_manager,
			train_size=1,
			validation_size=None,
			shuffle=False,
			shuffle_set_split=False,
			batch_size=None
		)
		training_plan = self._training_plan_cls(
			self.module, **plan_kwargs
		)

		if trainer_kwargs.get('early_stopping', False):
			# TODO: implement early stopping
			# Currently even if "early_stopping_monitor" is on training data
			# 'check_val_every_n_epoch' will still be set to one
			# in the Trainer class (scvi.train._train.Trainer L129)
			raise ValueError('Early stopping is not supported for SpatialANVI as there is no validation set.')

		runner = TrainRunner(
			self,
			training_plan=training_plan,
			data_splitter=data_splitter,
			max_epochs=max_epochs,
			accelerator=accelerator,
			devices=devices,
			**trainer_kwargs,
		)
		return runner()

	@classmethod
	def from_rna_model(
		cls,
		st_adata: AnnData,
		sc_model: SCANVI,
		spatial_loss: Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1,
		sp_loss_on: Literal['z', 'mean_z'] = 'mean_z',
		unfrozen: bool = False,
		freeze_dropout: bool = False,
		freeze_expression: bool = True,
		freeze_decoder_first_layer: bool = True,
		freeze_batchnorm_encoder: bool = True,
		freeze_batchnorm_decoder: bool = False,
		freeze_classifier: bool = True,
		**spanvae_kwargs,
	):
		"""Alternate constructor for exploiting a pre-trained model on RNA-seq data.

		Note that because of the dropout layer, even though the new instance is initialized
		with the same parameters as the pre-trained model, new_instance.get_latent_representation()
		may not return the same latent representation as the pre-trained model.

		Parameters
		----------
		st_adata
			registed anndata object
		sc_model
			pretrained SCANVI model
		"""

		# rechieve the parameters from the pretrained model
		init_params = sc_model.init_params_
		non_kwargs = deepcopy(init_params["non_kwargs"])
		kwargs = deepcopy(init_params["kwargs"])
		kwargs = {k: v for (i, j) in kwargs.items() for (k, v) in j.items()}
		for k, v in list({**non_kwargs, **kwargs}.items()):
			if k in spanvae_kwargs.keys():
				warnings.warn(
					f"Ignoring param '{k}' as it was already passed in to pretrained "
					f"SCVI model with value {v}.",
					UserWarning
				)
				del spanvae_kwargs[k]

		# overwrite the spatial loss parameters with the new ones
		if 'spatial_loss' in non_kwargs.keys():
			warnings.warn(
				"Overwriting the spatial_loss parameter of the pretrained model. "
				f"'spatial_loss': {non_kwargs['spatial_loss']} -> {spatial_loss}. "
				f"'lambda_spatial_loss': {non_kwargs['lambda_spatial_loss']} -> {lambda_spatial_loss}. "
				f"'sp_loss_on': {non_kwargs['sp_loss_on']} -> {sp_loss_on}.",
				UserWarning
			)
			del non_kwargs['spatial_loss']
			del non_kwargs['lambda_spatial_loss']
			del non_kwargs['sp_loss_on']

		# set up the anndata object
		registry = deepcopy(sc_model.adata_manager.registry)
		scanvi_setup_args = registry[_SETUP_ARGS_KEY]

		# prepare the query anndata and load the pretrained model
		st_adata_prep = SCANVI.prepare_query_anndata(st_adata, sc_model, inplace=False)
		cls.setup_anndata(
			st_adata_prep,
			source_registry=registry,
			extend_categories=True,
			allow_missing_labels=True,
			**scanvi_setup_args,
		)

		# initialize the new model
		sp_model = cls(
			st_adata_prep,
			spatial_loss=spatial_loss,
			lambda_spatial_loss=lambda_spatial_loss,
			sp_loss_on=sp_loss_on,
			**non_kwargs, **kwargs, **spanvae_kwargs
		)

		# copy the encoder and decoder state dict
		sc_state_dict = sc_model.module.state_dict()

		# model tweaking
		new_state_dict = sp_model.module.state_dict()
		for key, load_ten in sc_state_dict.items():
			new_ten = new_state_dict[key]
			if new_ten.size() == load_ten.size():
				continue
			# new categoricals changed size, need to pad the pretrained load_ten
			else:
				dim_diff = new_ten.size()[-1] - load_ten.size()[-1]
				fixed_ten = torch.cat([load_ten, new_ten[..., -dim_diff:]], dim=-1)
				sc_state_dict[key] = fixed_ten

		sp_model.module.load_state_dict(sc_state_dict)
		sp_model.module.eval()

		# sp_model = SCANVI.load_query_data(st_adata_prep, sc_model)
		# # change to the SpatialANVI class
		# sp_model.__class__ = cls
		# sp_model.n_genes = sp_model.summary_stats.n_vars
		# sp_model._model_summary_string = sp_model._model_summary_string.replace('ScanVI', 'SpatialANVI')
		# sp_model.init_params_['non_kwargs'].update({'spatial_loss': spatial_loss,
		# 											'lambda_spatial_loss': lambda_spatial_loss,
		# 											'sp_loss_on': sp_loss_on})

		# # switch to the spatial module
		# sp_model.module = SpatialANVI._module_cls(
		# 	spatial_loss=spatial_loss,
		# 	lambda_spatial_loss=lambda_spatial_loss,
		# 	sp_loss_on=sp_loss_on,
		# 	n_input=sp_model.n_genes,
		# 	n_batch=sp_model.summary_stats.n_batch,
		# 	n_labels=sp_model.summary_stats.n_labels - 1, # ignores unlabeled catgegory
		# 	n_continuous_cov=sp_model.summary_stats.get("n_extra_continuous_covs", 0),
		# 	**kwargs,
		# 	**non_kwargs,
		# 	**spanvae_kwargs,
		# )

		# freeze the pretrained model if neccessary
		set_params_online_update(
			sp_model.module,
			unfrozen=unfrozen,
			freeze_decoder_first_layer=freeze_decoder_first_layer,
			freeze_batchnorm_encoder=freeze_batchnorm_encoder,
			freeze_batchnorm_decoder=freeze_batchnorm_decoder,
			freeze_dropout=freeze_dropout,
			freeze_expression=freeze_expression,
			freeze_classifier=freeze_classifier,
		)
		sp_model.is_trained_ = False

		return sp_model



import warnings

import torch
from scvi import REGISTRY_KEYS, settings
from scvi.nn import one_hot, FCLayers

# SCVI.module.base._utils.py
def iterate(obj, func):
	"""Iterates over an object and applies a function to each element."""
	t = type(obj)
	if t is list or t is tuple:
		return t([iterate(o, func) for o in obj])
	else:
		return func(obj) if obj is not None else None

# SCVI.module.base._utils.py
def broadcast_labels(y, *o, n_broadcast=-1):
	"""Utility for the semi-supervised setting.

	If y is defined(labelled batch) then one-hot encode the labels (no broadcasting needed)
	If y is undefined (unlabelled batch) then generate all possible labels (and broadcast other arguments if not None)
	"""
	if not len(o):
		raise ValueError("Broadcast must have at least one reference argument")
	if y is None:
		ys = enumerate_discrete(o[0], n_broadcast)
		new_o = iterate(
			o,
			lambda x: x.repeat(n_broadcast, 1)
			if len(x.size()) == 2
			else x.repeat(n_broadcast),
		)
	else:
		ys = one_hot(y, n_broadcast)
		new_o = o
	return (ys,) + new_o

# SCVI.module.base._utils.py
def enumerate_discrete(x, y_dim):
	"""Enumerate discrete variables."""

	def batch(batch_size, label):
		labels = torch.ones(batch_size, 1, device=x.device, dtype=torch.long) * label
		return one_hot(labels, y_dim)

	batch_size = x.size(0)
	return torch.cat([batch(batch_size, i) for i in range(y_dim)])

# SCVI.module.base._utils.py
def get_max_epochs_heuristic(
	n_obs: int, epochs_cap: int = 400, decay_at_n_obs: int = 20000
) -> int:
	"""Compute a heuristic for the default number of maximum epochs.

	If `n_obs <= decay_at_n_obs`, the number of maximum epochs is set to
	`epochs_cap`. Otherwise, the number of maximum epochs decays according to
	`(decay_at_n_obs / n_obs) * epochs_cap`, with a minimum of 1.

	Parameters
	----------
	n_obs
		The number of observations in the dataset.
	epochs_cap
		The maximum number of epochs for the heuristic.
	decay_at_n_obs
		The number of observations at which the heuristic starts decaying.

	Returns
	-------
	`int`
		A heuristic for the default number of maximum epochs.
	"""
	max_epochs = min(round((decay_at_n_obs / n_obs) * epochs_cap), epochs_cap)
	max_epochs = max(max_epochs, 1)

	if max_epochs == 1:
		warnings.warn(
			"The default number of maximum epochs has been set to 1 due to the large"
			"number of observations. Pass in `max_epochs` to the `train` function in "
			"order to override this behavior.",
			UserWarning,
			stacklevel=settings.warnings_stacklevel,
		)

	return max_epochs

# SCVI.model.base._archesmixins.py
def set_params_online_update(
	module,
	unfrozen,
	freeze_decoder_first_layer,
	freeze_batchnorm_encoder,
	freeze_batchnorm_decoder,
	freeze_dropout,
	freeze_expression,
	freeze_classifier,
):
	"""Freeze parts of network for scArches."""
	# do nothing if unfrozen
	if unfrozen:
		return

	mod_inference_mode = {"encoder_z2_z1", "decoder_z1_z2"}
	mod_no_hooks_yes_grad = {"l_encoder"}
	if not freeze_classifier:
		mod_no_hooks_yes_grad.add("classifier")
	parameters_yes_grad = {"background_pro_alpha", "background_pro_log_beta"}

	def no_hook_cond(key):
		one = (not freeze_expression) and "encoder" in key
		two = (not freeze_decoder_first_layer) and "px_decoder" in key
		return one or two

	def requires_grad(key):
		mod_name = key.split(".")[0]
		# linear weights and bias that need grad
		one = "fc_layers" in key and ".0." in key and mod_name not in mod_inference_mode
		# modules that need grad
		two = mod_name in mod_no_hooks_yes_grad
		three = sum([p in key for p in parameters_yes_grad]) > 0
		# batch norm option
		four = (
			"fc_layers" in key
			and ".1." in key
			and "encoder" in key
			and (not freeze_batchnorm_encoder)
		)
		five = (
			"fc_layers" in key
			and ".1." in key
			and "decoder" in key
			and (not freeze_batchnorm_decoder)
		)
		if one or two or three or four or five:
			return True
		else:
			return False

	for key, mod in module.named_modules():
		# skip over protected modules
		if key.split(".")[0] in mod_no_hooks_yes_grad:
			continue
		if isinstance(mod, FCLayers):
			hook_first_layer = False if no_hook_cond(key) else True
			mod.set_online_update_hooks(hook_first_layer)
		if isinstance(mod, torch.nn.Dropout):
			if freeze_dropout:
				mod.p = 0
		# momentum freezes the running stats of batchnorm
		freeze_batchnorm = ("decoder" in key and freeze_batchnorm_decoder) or (
			"encoder" in key and freeze_batchnorm_encoder
		)
		if isinstance(mod, torch.nn.BatchNorm1d) and freeze_batchnorm:
			mod.momentum = 0

	for key, par in module.named_parameters():
		if requires_grad(key):
			par.requires_grad = True
		else:
			par.requires_grad = False



from typing import List, Literal, Optional, Union
from copy import deepcopy
from timeit import default_timer as timer
import warnings
import inspect

import torch
from torch.distributions import kl_divergence as kl
from scvi import REGISTRY_KEYS
from scvi.module.base import LossOutput
from scvi.module import VAE
from scvi.model import SCVI

from anndata import AnnData
from scvi.dataloaders import DeviceBackedDataSplitter
from scvi.utils._docstrings import devices_dsp
from scvi.data._constants import _SETUP_ARGS_KEY

from smoother import SpatialLoss
from ._utils import set_params_online_update

class SPVAE(VAE):
	"""Add spatial loss to the latent representation in the VAE model.

	Parameters:
		spatial_loss: Spatial loss to apply on the latent representation. If None, no spatial loss.
		lambda_spatial_loss: Weight of the spatial loss.
		sp_loss_on: Whether to apply the spatial loss on the sampled latent representation ('z') or
			on the encoded mean of the latent distribution ('mean_z').
	"""

	def __init__(
		self,
		n_input: int,
		spatial_loss: Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1,
		sp_loss_on: Literal['z', 'mean_z'] = 'mean_z',
		n_hidden: int = 128,
		n_latent: int = 10,
		n_layers: int = 1,
		dropout_rate: float = 0.1,
		dispersion: Literal["gene", "gene-batch", "gene-label", "gene-cell"] = "gene",
		gene_likelihood: Literal["zinb", "nb", "poisson"] = "nb",
		latent_distribution: Literal["normal", "ln"] = "normal",
		**model_kwargs
	):
		# initialize the VAE model
		super().__init__(
			n_input=n_input,
			n_hidden=n_hidden,
			n_latent=n_latent,
			n_layers=n_layers,
			dropout_rate=dropout_rate,
			dispersion=dispersion,
			gene_likelihood=gene_likelihood,
			latent_distribution=latent_distribution,
			**model_kwargs
		)

		# spatial loss
		self.spatial_loss = spatial_loss
		self.l_sp_loss = lambda_spatial_loss
		self.sp_loss_on = sp_loss_on

	def loss(
		self,
		tensors,
		inference_outputs,
		generative_outputs,
		kl_weight: float = 1.0,
	):
		# calculate the original VAE loss
		x = tensors[REGISTRY_KEYS.X_KEY]
		kl_divergence_z = kl(inference_outputs["qz"], generative_outputs["pz"]).sum(
			dim=-1
		)
		if not self.use_observed_lib_size:
			kl_divergence_l = kl(
				inference_outputs["ql"],
				generative_outputs["pl"],
			).sum(dim=1)
		else:
			kl_divergence_l = torch.tensor(0.0, device=x.device)

		reconst_loss = -generative_outputs["px"].log_prob(x).sum(-1)

		kl_local_for_warmup = kl_divergence_z
		kl_local_no_warmup = kl_divergence_l

		weighted_kl_local = kl_weight * kl_local_for_warmup + kl_local_no_warmup

		# add the spatial loss
		if self.spatial_loss is not None and self.l_sp_loss > 0.0:
			if self.sp_loss_on == 'z':
				sp_loss = self.spatial_loss(inference_outputs['z'].T, normalize = False)
			elif self.sp_loss_on == 'mean_z':
				sp_loss = self.spatial_loss(inference_outputs['qz'].loc.T, normalize = False)
			else:
				raise NotImplementedError('Currently the spatial loss can only be applied on z or mean_z.')

			sp_loss *= self.l_sp_loss / self.n_latent
		else:
			sp_loss = torch.tensor(0.0)

		loss = torch.mean(reconst_loss + weighted_kl_local + sp_loss)

		kl_local = {
			"kl_divergence_l": kl_divergence_l,
			"kl_divergence_z": kl_divergence_z,
		}
		return LossOutput(
			loss=loss, reconstruction_loss=reconst_loss, kl_local=kl_local, kl_global=sp_loss
		)

	@classmethod
	def from_vae(
		cls,
		vae_module: VAE,
		spatial_loss: Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1,
		sp_loss_on: Literal['z', 'mean_z'] = 'mean_z'
	):

		# copy the VAE module
		spvae = deepcopy(vae_module)

		# switch to the spatial module
		spvae.__class__ = cls
		spvae.spatial_loss = spatial_loss
		spvae.l_sp_loss = lambda_spatial_loss
		spvae.sp_loss_on = sp_loss_on

		return spvae


class SpatialVAE(SCVI):
	"""Spatially-aware Variational Autoencoder model.
	"""

	_data_splitter_cls = DeviceBackedDataSplitter
	_module_cls = SPVAE

	def __init__(
		self,
		st_adata: AnnData,
		spatial_loss: Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1,
		sp_loss_on: Literal['z', 'mean_z'] = 'mean_z',
		n_hidden: int = 128,
		n_latent: int = 10,
		n_layers: int = 1,
		dropout_rate: float = 0.0,
		dispersion: Literal["gene", "gene-batch", "gene-label", "gene-cell"] = "gene",
		gene_likelihood: Literal["zinb", "nb", "poisson"] = "nb",
		latent_distribution: Literal["normal", "ln"] = "normal",
		**model_kwargs,
	):
		super().__init__(
			st_adata,
			n_hidden=n_hidden,
			n_latent=n_latent,
			n_layers=n_layers,
			dropout_rate=dropout_rate,
			dispersion=dispersion,
			gene_likelihood=gene_likelihood,
			latent_distribution=latent_distribution,
			**model_kwargs,
		)

		self.module = self._module_cls.from_vae(
			self.module,
			spatial_loss=spatial_loss,
			lambda_spatial_loss=lambda_spatial_loss,
			sp_loss_on=sp_loss_on
		)

		self._model_summary_string = self._model_summary_string.replace('SCVI', 'SpatialVAE')
		self.init_params_ = self._get_init_params(locals())
		self.dr_logs = {'elapsed_time': None, 'total_loss': [],
						'recon_loss': [], 'spatial_loss': []}

	@devices_dsp.dedent
	def train(
		self,
		max_epochs: int = 400,
		lr: float = 0.01,
		accelerator: str = "auto",
		devices: Union[int, List[int], str] = "auto",
		plan_kwargs: Optional[dict] = None,
		**kwargs,
	):
		"""Trains the model without mini-batch.
		"""
		update_dict = {
			"lr": lr,
		}
		if plan_kwargs is not None:
			plan_kwargs.update(update_dict)
		else:
			plan_kwargs = update_dict

		t_start = timer()

		if kwargs.get('early_stopping', False):
			# TODO: implement early stopping
			# Currently even if "early_stopping_monitor" is on training data
			# 'check_val_every_n_epoch' will still be set to one
			# in the Trainer class (scvi.train._train.Trainer L129)
			raise ValueError('Early stopping is not supported for SpatialVAE as there is no validation set.')

		# fit the model with all spots (no mini-batch)
		super().train(
			max_epochs=max_epochs,
			accelerator=accelerator,
			devices=devices,
			train_size=1,
			validation_size=None,
			shuffle_set_split=False,
			batch_size=None,
			plan_kwargs=plan_kwargs,
			**kwargs,
		)
		t_end = timer()
		self.dr_logs['elapsed_time'] = t_end - t_start
		self.dr_logs['total_loss'] = self.trainer.logger.history['elbo_train']
		self.dr_logs['recon_loss'] = self.trainer.logger.history['reconstruction_loss_train']
		self.dr_logs['spatial_loss'] = self.trainer.logger.history['kl_global_train']


	@classmethod
	def from_rna_model(
		cls,
		st_adata: AnnData,
		sc_model: SCVI,
		spatial_loss: Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1,
		sp_loss_on: Literal['z', 'mean_z'] = 'mean_z',
		unfrozen: bool = False,
		freeze_dropout: bool = False,
		freeze_expression: bool = True,
		freeze_decoder_first_layer: bool = True,
		freeze_batchnorm_encoder: bool = True,
		freeze_batchnorm_decoder: bool = False,
		freeze_classifier: bool = True,
		**spvae_kwargs,
	):
		"""Alternate constructor for exploiting a pre-trained model on RNA-seq data.

		Note that because of the dropout layer, even though the new instance is initialized
		with the same parameters as the pre-trained model, new_instance.get_latent_representation()
		may not return the same latent representation as the pre-trained model.

		Parameters
		----------
		st_adata
			registed anndata object
		sc_model
			pretrained SCVI model
		"""
		# rechieve the parameters from the pretrained model
		init_params = sc_model.init_params_
		non_kwargs = deepcopy(init_params["non_kwargs"])
		kwargs = deepcopy(init_params["kwargs"])
		kwargs = {k: v for (i, j) in kwargs.items() for (k, v) in j.items()}
		for k, v in list({**non_kwargs, **kwargs}.items()):
			if k in spvae_kwargs.keys():
				warnings.warn(
					f"Ignoring param '{k}' as it was already passed in to pretrained "
					f"SCVI model with value {v}.",
					UserWarning
				)
				del spvae_kwargs[k]

		# overwrite the spatial loss parameters with the new ones
		if 'spatial_loss' in non_kwargs.keys():
			warnings.warn(
				"Overwriting the spatial_loss parameter of the pretrained model. "
				f"'spatial_loss': {non_kwargs['spatial_loss']} -> {spatial_loss}. "
				f"'lambda_spatial_loss': {non_kwargs['lambda_spatial_loss']} -> {lambda_spatial_loss}. "
				f"'sp_loss_on': {non_kwargs['sp_loss_on']} -> {sp_loss_on}.",
				UserWarning
			)
			del non_kwargs['spatial_loss']
			del non_kwargs['lambda_spatial_loss']
			del non_kwargs['sp_loss_on']

		# set up the anndata object
		registry = deepcopy(sc_model.adata_manager.registry)
		scvi_setup_args = registry[_SETUP_ARGS_KEY]
		valid_setup_args = inspect.getfullargspec(cls.setup_anndata).args
		for k in list(scvi_setup_args.keys()):
			if k not in valid_setup_args:
				warnings.warn(
					f"Argument '{k}' in the pretrained model is not valid for {cls.__name__}.setup_anndata()."
					" Will be ignored.",
					UserWarning
				)
				del scvi_setup_args[k]

		# prepare the query anndata
		st_adata_prep = SCVI.prepare_query_anndata(st_adata, sc_model, inplace=False)
		cls.setup_anndata(
			st_adata_prep,
			source_registry=registry,
			extend_categories=True,
			allow_missing_labels=True,
			**scvi_setup_args,
		)

		# initialize the new model
		sp_model = cls(
			st_adata_prep,
			spatial_loss=spatial_loss,
			lambda_spatial_loss=lambda_spatial_loss,
			sp_loss_on=sp_loss_on,
			**non_kwargs, **kwargs, **spvae_kwargs
		)

		# copy the encoder and decoder state dict
		sc_state_dict = sc_model.module.state_dict()

		# model tweaking
		new_state_dict = sp_model.module.state_dict()
		for key, load_ten in sc_state_dict.items():
			new_ten = new_state_dict[key]
			if new_ten.size() == load_ten.size():
				continue
			# new categoricals changed size, need to pad the pretrained load_ten
			else:
				dim_diff = new_ten.size()[-1] - load_ten.size()[-1]
				fixed_ten = torch.cat([load_ten, new_ten[..., -dim_diff:]], dim=-1)
				sc_state_dict[key] = fixed_ten

		sp_model.module.load_state_dict(sc_state_dict)
		sp_model.module.eval()

		# freeze the pretrained model if neccessary
		set_params_online_update(
			sp_model.module,
			unfrozen=unfrozen,
			freeze_decoder_first_layer=freeze_decoder_first_layer,
			freeze_batchnorm_encoder=freeze_batchnorm_encoder,
			freeze_batchnorm_decoder=freeze_batchnorm_decoder,
			freeze_dropout=freeze_dropout,
			freeze_expression=freeze_expression,
			freeze_classifier=freeze_classifier,
		)
		sp_model.is_trained_ = False

		return sp_model



"""
Extract spatial patterns from a given histology image
"""

import warnings
import math
import operator
from tqdm import tqdm
import pandas as pd
import numpy as np

try:
	from squidpy.im import ImageContainer
except ImportError:
	warnings.warn("Package 'Squidpy' not installed.", category=ImportWarning)
	ImageContainer = None

def get_features_matrix(crop : ImageContainer, layer):
	'''
	Extract features from a given crop

	Args:
		crop: crop of a image
		layer: Image layer in crop that should be processed

	Returns:
		df_1: the extracted features at this crop
	'''
	text = crop.features_texture(layer = layer)
	hist = crop.features_histogram(layer = layer)
	summ = crop.features_summary(layer = layer)
	dict_1 = text | hist| summ
	df_1 = pd.DataFrame(dict_1.items()).transpose()
	df_1.columns = df_1.iloc[0]
	return df_1.drop(df_1.index[0])

def extract_features_from_image(image : ImageContainer, crop_size, layer):
	'''
	Extract features from a given image

	Args:
		image: squidpy ImageContainer
		crop_size: the # of crops wanted [n_row_crops, n_col_crops]
		layer: Image layer in image that should be processed

	Returns:
		features_df: the extracted features at all crops
	'''
	features_df = pd.DataFrame()
	total = int(image.shape[0]/crop_size[0]) * int((image.shape[1]/crop_size[1]))
	for crop in tqdm(image.generate_equal_crops(size = crop_size), total = total):
		features_df_1 = get_features_matrix(crop, layer)
		features_df = pd.concat((features_df, features_df_1), axis=0)
	features_df.index = [f"location_{i}" for i in range(len(features_df))]
	return features_df

def cal_membership_value(membership_mat, # membership matrix n_location x n_cluster
						 feature_df, # features matrix n_location x n_feature
						 cluster_centers, # center point for clusters n_cluster x n_feature
						 m): # Calculating the membership value
	'''
	Calculate fuzzy-c-means membership values

	Args:
		membership_mat: membership matrix n_location x n_cluster
		feature_df: features matrix n_location x n_feature
		cluster_centers: center point for clusters n_cluster x n_feature
		m: m value required by FCM

	Returns:
		membership_mat: updated membership matrix n_location x n_cluster
	'''
	p = float(2/(m-1))
	n = membership_mat.shape[0]
	k = cluster_centers.shape[0]
	for i in range(n):
		x = list(feature_df.iloc[i])
		distances = [np.linalg.norm(np.array(list(map(operator.sub, x, cluster_centers[j]))))
               		 for j in range(k)]
		for j in range(k):
			den = sum([math.pow(float(distances[j]/distances[c]), p) for c in range(k)])
			membership_mat[i][j] = float(1/den)
	return membership_mat



from typing import List, Literal, Optional, Union
from copy import deepcopy
from timeit import default_timer as timer
import warnings

import numpy as np
import torch
import torch.nn as nn
from torch.optim.lr_scheduler import ReduceLROnPlateau

from anndata import AnnData
from scvi.nn import FCLayers

from smoother.losses import SpatialLoss


class AutoEncoderClass(nn.Module):
	"""Abstract AutoEncoder class.

	To mimic the behavior of PCA, the overall objective contains three losses:
		1. Reconstruction loss: reconstruction error of the observed data.
		2. Orthogonality loss: the loss of orthogonality on the hidden embeddings.
		3. Spatial loss: the spatial loss on the hidden embeddings.
	"""
	def __init__(self,
		n_feature: int,
		n_latent: int = 10,
		spatial_loss :Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1,
		lambda_orth_loss = 0.1
	) -> None:
		super().__init__()

		self.n_latent = n_latent # dimension of the latent space
		self.n_feature = n_feature # num_feature to project

		# store data to project
		self._data = None

		# store spatial loss
		self.spatial_loss = spatial_loss
		self.l_sp_loss = lambda_spatial_loss

		# store orthogonality loss
		self.l_orth_loss = lambda_orth_loss

		# configs and logs for dimension reduction
		self.dr_configs = {}
		self.dr_logs = {'elapsed_time': None, 'total_loss': [],
						'recon_loss': [], 'orth_loss': [],
						'spatial_loss': []}

	def load_adata(self, adata: AnnData, layer: Optional[str] = None):
		"""Load data to project.

		Args:
			adata (AnnData): data to project.
			layer (str): layer to project. If None, use adata.X.
		"""
		assert adata.shape[1] == self.n_feature, "The new adata has different number of features!"

		if layer is None:
			self._data = torch.tensor(adata.X.toarray().T) # n_feature x n_sample
		else:
			data = adata.layers[layer]
			if not isinstance(data, np.ndarray):
				data = data.toarray()
			self._data = torch.tensor(data.T) # n_feature x n_sample

		self._data = self._data.float()

	def forward(self, x):
		return self.decode(self.encode(x))

	def encode(self, x):
		return NotImplementedError

	def decode(self, x_enc):
		return NotImplementedError

	def get_latent_representation(self):
		"""Get the latent representation of the loaded data."""
		return self.encode(self._data).detach().numpy()

	def get_recon_loss(self):
		"""Get the reconstruction loss."""
		raise NotImplementedError

	def get_orth_loss(self):
		"""Get the orthogonality loss (weighted by lambda_orth_loss)."""
		raise NotImplementedError

	def get_sp_loss(self):
		"""Get the spatial loss (weighted by lambda_spatial_loss)."""
		raise NotImplementedError

	def reduce(self, lr = 1e-2, max_epochs = 1000, patience = 10, tol = 1e-5,
			   optimizer = 'SGD', verbose = True, quite = False, clear_logs = True):

		self.dr_configs = {
			'lr':lr, 'max_epochs':max_epochs, 'patience':patience, 'tol':tol,
			'optimizer':optimizer, 'verbose':verbose, 'quite':quite,
   			'clear_logs':clear_logs, 'return_model':False
		}
		return AutoEncoderClass._reduce_dim_ae(ae_model=self, **self.dr_configs)

	@classmethod
	def _reduce_dim_ae(
		cls, ae_model,
		lr = 1e-2, max_epochs = 1000, patience = 10, tol = 1e-5,
		optimizer = 'SGD', verbose = True, quite = False,
		clear_logs = False, return_model = False
	):
		"""Dimension reduction using auto-encoders.

		Two additional losses are added in addition to the reconstruction loss:
			1. Orthogonality loss: the loss of the orthogonality of the hidden embedding.
			2. Spatial loss: the spatial loss on the hidden embeddings.

		Args:
			ae_model (AutoEncoder): The autoencoder model.
			lr (float): The learning rate.
			max_epochs (int): The maximum number of epochs.
			patience (int): The patience for early stopping.
			tol (float): The tolerated convergence error.
			optimizer (str): The optimizer to be used. Can be 'SGD' or 'Adam'.
			verbose (bool): If True, print out loss while training.
			quite (bool): If True, no output printed.
			clear_logs (bool): If True, clear the logs in the autoencoder model.
			return_model (bool): If True, return the trained autoencoder model.
		Returns:
			ae_model (AutoEncoder): The trained autoencoder model.
		"""

		# start timer
		t_start = timer()

		# set optimizer
		assert optimizer in ['SGD', 'Adam']
		if optimizer == 'Adam':
			optimizer = torch.optim.Adam(ae_model.parameters(), lr=lr)
		else:
			optimizer = torch.optim.SGD(ae_model.parameters(), lr=lr, momentum=0.9)

		# set scheduler
		scheduler = ReduceLROnPlateau(
			optimizer,
			patience = int(patience / 4) + 1,
			factor = 0.1,
			verbose = verbose
		)

		# clear logs
		if clear_logs:
			ae_model.dr_logs = {'elapsed_time': None, 'total_loss': [],
							'recon_loss': [], 'orth_loss': [],
							'spatial_loss': []}

		# start training
		ae_model.train()

		# set iteration limits
		epoch, prev_loss, d_loss = 0, 100000.0, 100000.0
		max_epochs = 10000 if max_epochs == -1 else max_epochs
		patience = patience if patience > 0 else 1

		while epoch < max_epochs and d_loss >= tol or (d_loss < 0 and patience > 0):

			optimizer.zero_grad()

			# calculate reconstruction error
			recon_loss = ae_model.get_recon_loss()
			loss = recon_loss
			recon_loss = recon_loss.detach().item()

			# calculate orthogonality loss
			if ae_model.l_orth_loss > 0:
				orth_loss = ae_model.get_orth_loss()
				loss += orth_loss
				orth_loss = orth_loss.detach().item()
			else:
				orth_loss = 0.0

			# calculate spatial loss
			if ae_model.spatial_loss is not None and ae_model.l_sp_loss > 0:
				sp_loss = ae_model.get_sp_loss()
				loss += sp_loss
				sp_loss = sp_loss.detach().item()
			else:
				sp_loss = 0.0

			# backpropagate and update weights
			loss.backward()
			optimizer.step()

			# check convergence
			epoch += 1
			d_loss = prev_loss - loss.detach().item()
			if d_loss < 0: # if loss increases
				patience -= 1

			prev_loss = loss.detach().item()

			# update learning rate
			scheduler.step(prev_loss)

			if (verbose and not quite) and epoch % 10 == 0:
				print(f'Epoch {epoch}. Loss: (total) {prev_loss:.4f}. (recon) {recon_loss:.4f}. '
					f'(orth) {orth_loss:.4f}. (spatial) {sp_loss:.4f}.')

			# store losses to the log
			ae_model.dr_logs['total_loss'].append(prev_loss)
			ae_model.dr_logs['recon_loss'].append(recon_loss)
			ae_model.dr_logs['spatial_loss'].append(sp_loss)
			ae_model.dr_logs['orth_loss'].append(orth_loss)

		t_end = timer()
		ae_model.dr_logs['elapsed_time'] = t_end - t_start

		if not quite: # print final message
			print(f'=== Time {t_end - t_start : .2f}s. Total epoch {epoch}. '
					f'Final Loss: (total) {prev_loss:.4f}. (recon) {recon_loss:.4f}. '
						f'(orth) {orth_loss:.4f}. (spatial) {sp_loss:.4f}.')

		if d_loss >= 1e-5:
			warnings.warn("Fail to converge. Try to increase 'max_epochs'.")

		if return_model:
			return ae_model


class SpatialAutoEncoder(AutoEncoderClass):
	"""Generic spatial auto-encoder."""
	def __init__(
		self,
		adata: AnnData,
		layer: Optional[str] = None,
		spatial_loss :Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1,
		lambda_orth_loss = 0.1,
		recon_loss_mode: Literal['mse', 'poisson'] = 'poisson',
		n_layers: int = 1,
		n_hidden: int = 128,
		n_latent: int = 10,
		dropout_rate: float = 0.0,
		use_batch_norm: bool = True,
		use_activation: bool = True,
		activation_fn: Optional[nn.Module] = nn.ReLU,
		use_bias = True
	) -> None:
		# initialize the model
		n_feature = adata.shape[1]
		super().__init__(
			n_feature = n_feature,
			n_latent = n_latent,
			spatial_loss = spatial_loss,
			lambda_spatial_loss = lambda_spatial_loss,
			lambda_orth_loss = lambda_orth_loss
		)

		# store set up arguments
		self.setup_args = {
			'lambda_orth_loss': lambda_orth_loss,
			'recon_loss_mode': recon_loss_mode,
			'n_layers': n_layers,
			'n_hidden': n_hidden,
			'n_latent': n_latent,
			'dropout_rate': dropout_rate,
			'use_batch_norm': use_batch_norm,
			'use_activation': use_activation,
			'activation_fn': activation_fn,
			'use_bias': use_bias
		}

		# reconstruction loss function
		if recon_loss_mode	== 'mse':
			self._recon_loss_fn = nn.MSELoss(reduction='mean')
		elif recon_loss_mode == 'poisson':
			self._recon_loss_fn = nn.PoissonNLLLoss(log_input=False, reduction='mean')
		else:
			raise NotImplementedError

		# load data
		self.load_adata(adata, layer)
		self.lib_size = self._data.sum(dim=0) # per-spot library size

		# initialize the encoder
		self.encoder = FCLayers(
			n_in = n_feature,
			n_out = n_latent,
			n_layers = n_layers,
			n_hidden = n_hidden,
			dropout_rate = dropout_rate,
			use_batch_norm = use_batch_norm,
			use_activation = use_activation,
			activation_fn = activation_fn,
			bias = use_bias
		)

		# initialize the decoder
		self.decoder = nn.Sequential(
  			FCLayers(
				n_in = n_latent,
				n_out = n_hidden,
				n_layers = n_layers,
				n_hidden = n_hidden,
				dropout_rate = 0,
				use_batch_norm = use_batch_norm,
				use_activation = use_activation,
				activation_fn = activation_fn,
				bias = use_bias
			),
			nn.Linear(n_hidden, n_feature, bias=use_bias),
			nn.Softplus()
		)

	def encode(self, x):
		"""Project the input to the hidden space.

		Args:
			x (2D tensor): The input matrix, num_gene x num_spot.

		Returns:
			x_enc (2D tensor): The hidden embedding, num_spot x num_hidden.
		"""
		return self.encoder(x.T)

	def decode(self, x_enc):
		"""Project the hidden embedding back to the input space.

		Args:
			x_enc (2D tensor): The hidden embedding, num_spot x num_hidden.

		Returns:
			x_dec (2D tensor): The decoded matrix, num_gene x num_spot.
		"""
		x_dec = self.decoder(x_enc).T

		return x_dec

	def get_recon_loss(self):
		x = self._data # num_gene x num_spot
		y = self(x) # num_gene x num_spot
		return self._recon_loss_fn(y, x)

	def get_orth_loss(self):
		x = self._data # num_gene x num_spot

		if self.l_orth_loss > 0:
			enc = self.encode(x) # num_spot x num_latent
			enc_norm = enc.norm(dim=0, keepdim=True) + 1e-8
			# loss to make sure the latent embedding is orthogonal
			orth_loss = torch.norm(
				(enc / enc_norm).T @ (enc / enc_norm) - torch.eye(enc.shape[1])
			)
			# add l2 loss to constrain the scale of the latent embedding
			orth_loss = self.l_orth_loss * orth_loss

			return orth_loss
		else:
			return 0.0

	def get_sp_loss(self):
		x = self._data # num_gene x num_spot

		if self.spatial_loss is not None and self.l_sp_loss > 0:
			enc = self.encode(x) # num_spot x num_latent
			sp_loss = self.l_sp_loss * self.spatial_loss(enc.T, normalize = True)

			return sp_loss
		else:
			return 0.0

	@classmethod
	def from_rna_model(
		cls,
		rna_model,
		st_adata: AnnData,
		layer: Optional[str] = None,
		spatial_loss :Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1
	):
		"""Initialize a spatial model from a pre-trained RNA model."""
		model_args = deepcopy(rna_model.setup_args)
		sp_model = cls(
			adata = st_adata,
			layer = layer,
			spatial_loss = spatial_loss,
			lambda_spatial_loss = lambda_spatial_loss,
			**model_args
		)

		# copy the encoder and decoder state dict
		sp_model.load_state_dict(rna_model.state_dict(), strict=False)

		return sp_model



from typing import List, Literal, Optional, Union
from timeit import default_timer as timer
import warnings
import numpy as np
import torch
import torch.nn as nn
from torch.optim.lr_scheduler import ReduceLROnPlateau

from anndata import AnnData
from smoother import SpatialLoss

class SpatialPCA(nn.Module):
	"""Solving PCA using stochastic gradient descent."""
	def __init__(
		self,
		adata: AnnData,
		layer: Optional[str] = None,
		n_latent: int = 10,
		spatial_loss :Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1
	) -> None:
		# initialize the PCA model.
		super().__init__()

		self.n_latent = n_latent # num_pc
		self.n_feature = adata.shape[1] # num_feature to project

		# store data to project
		self._data = None
		self.load_adata(adata, layer)

		# store spatial loss
		self.spatial_loss = spatial_loss
		self.l_sp_loss = lambda_spatial_loss

		# initialize PC basis matrix (num_feature x num_pc)
		U_init = self._gram_schmidt(nn.init.uniform_(torch.zeros(self.n_feature, self.n_latent)))
		self.U = nn.parameter.Parameter(U_init, requires_grad = True)

		# configs and logs for dimension reduction
		self.dr_configs = {}
		self.dr_logs = {'elapsed_time': None, 'total_loss': [],
					 'recon_loss': [], 'spatial_loss': []}

	def load_adata(self, adata: AnnData, layer: Optional[str] = None):
		"""Load data to project.

		Args:
			adata (AnnData): data to project.
			layer (str): layer to project. If None, use adata.X.
		"""
		assert adata.shape[1] == self.n_feature, "The new adata has different number of features!"

		if layer is None:
			self._data = torch.tensor(adata.X.toarray().T) # n_feature x n_sample
		else:
			data = adata.layers[layer]
			if not isinstance(data, np.ndarray):
				data = data.toarray()
			self._data = torch.tensor(data.T) # n_feature x n_sample

		self._data = self._data.float()

	def _gram_schmidt(self, U):
		"""Project the PC basis matrix to the feasible space.

		Args:
			U (2D tensor): PC basis matrix, num_feature x num_pc.
		"""
		# U will be orthogonal after transformation
		return torch.linalg.qr(U, mode = 'reduced')[0]

	def forward(self, x):
		"""Project x to the lower PC space.

		Args:
			x (2D tensor): data to project, num_feature x num_sample.
		"""
		return self.U.T @ x # num_pc x num_sample

	def _init_with_svd(self, x):
		"""Initialize model with svd solution.

		Args:
			x (2D tensor): data to project, num_feature x num_sample.
		"""
		with torch.no_grad():
			_, _, V = torch.svd_lowrank(x.T, self.n_latent)
			self.U.copy_(V) # num_feature x num_pc

	def get_latent_representation(self):
		"""Project loaded data to the lower PC space and return the latent representation.
		"""
		return self(self._data).T.detach().numpy() # num_sample x num_pc

	def reduce(self, lr = 1.0, max_epochs = 1000, patience = 10, tol = 1e-5,
			   init_with_svd = False, verbose = True, quite = False, clear_logs = True):
		"""Reduce the dimension of the expression matrix.

		Args:
			lr (float): The learning rate.
			max_epochs (int): The maximum number of epochs.
			patience (int): The patience for early stopping.
			tol (float): The tolerated convergence error.
			init_with_svd (bool): Whether to initialize with analytical solution calculated
				using `torch.svd_lowrank()`.
			verbose (bool): If True, print out loss while training.
			quite (bool): If True, no output printed.
			clear_logs (bool): If True, clear logs before training.
		"""
		self.dr_configs = {
			'lr':lr, 'max_epochs':max_epochs, 'patience':patience, 'tol':tol,
			'init_with_svd':init_with_svd
		}

		# start timer
		t_start = timer()

		optimizer = torch.optim.SGD(self.parameters(), lr=lr, momentum=0.9)
		scheduler = ReduceLROnPlateau(
			optimizer,
			patience = int(patience / 4) + 1,
			factor = 0.1,
			verbose = verbose
		)

		if init_with_svd:
			self._init_with_svd(self._data)

		if clear_logs:
			self.dr_logs = {'elapsed_time': None, 'total_loss': [],
							'recon_loss': [], 'spatial_loss': []}

		# start training
		self.train()

		# set iteration limits
		epoch, prev_loss, d_loss = 0, 100000.0, 100000.0
		max_epochs = 10000 if max_epochs == -1 else max_epochs
		patience = patience if patience > 0 else 1

		while epoch < max_epochs and d_loss >= tol or (d_loss < 0 and patience > 0):

			# set loss to 0
			recon_loss, sp_loss = torch.tensor(0), torch.tensor(0)

			optimizer.zero_grad()

			loss = 0

			# project data to hidden subspace
			hidden = self(self._data) # k x num_spot

			# calculate reconstruction error
			# here we want to maximize the variance in the subspace
			recon_loss = - torch.norm(hidden, dim = 0).pow(2).sum() / (hidden.shape[0] * hidden.shape[1])
			loss += recon_loss

			# add spatial loss on U.T @ x
			if self.spatial_loss is not None and self.l_sp_loss > 0:
				sp_loss = self.l_sp_loss * self.spatial_loss(hidden, normalize = True)
				loss += sp_loss

			# backpropagate and update weights
			loss.backward()
			optimizer.step()

			# project U to the feasible space
			with torch.no_grad():
				self.U.copy_(self._gram_schmidt(self.U))

			# check convergence
			epoch += 1
			d_loss = prev_loss - loss.detach().item()
			if d_loss < 0: # if loss increases
				patience -= 1

			prev_loss = loss.detach().item()
			recon_loss = recon_loss.detach().item()
			sp_loss = sp_loss.detach().item()

			# update learning rate
			scheduler.step(prev_loss)

			# store losses to the log
			self.dr_logs['total_loss'].append(prev_loss)
			self.dr_logs['recon_loss'].append(recon_loss)
			self.dr_logs['spatial_loss'].append(sp_loss)

			if (verbose and not quite) and epoch % 10 == 0:
				print(f'Epoch {epoch}. Loss: (total) {prev_loss:.4f}. (recon) {recon_loss:.4f}. '
					  f'(spatial) {sp_loss:.4f}.')

		t_end = timer()
		self.dr_logs['elapsed_time'] = t_end - t_start

		if not quite: # print final message
			print(f'=== Time {t_end - t_start : .2f}s. Total epoch {epoch}. '
				  f'Final Loss: (total) {prev_loss:.4f}. (recon) {recon_loss:.4f}. '
					  f'(spatial) {sp_loss:.4f}.')

		if d_loss >= tol:
			warnings.warn("Fail to converge. Try to increase 'max_epochs'.")

	@classmethod
	def from_rna_model(
		cls,
		rna_model,
		st_adata: AnnData,
		layer: Optional[str] = None,
		spatial_loss :Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1
	):
		"""Initialize a spatial model from a pre-trained RNA model."""
		pca_sp = cls(
			adata = st_adata,
			layer = layer,
			n_latent = rna_model.n_latent,
			spatial_loss = spatial_loss,
			lambda_spatial_loss = lambda_spatial_loss
		)

		# copy the PCA loadings from the RNA model
		with torch.no_grad():
			pca_sp.U.copy_(rna_model.U)

		return pca_sp

	@classmethod
	def from_scanpy(
		cls,
		adata: AnnData,
		layer: Optional[str] = None,
		spatial_loss :Optional[SpatialLoss] = None,
		lambda_spatial_loss = 0.1
	):
		"""Initialize a spatial model from a pre-trained RNA model."""
		assert 'PCs' in adata.varm.keys(), "Please run 'sc.tl.pca()' first."
		n_latent = adata.varm['PCs'].shape[1]
		pca_sp = cls(
			adata = adata,
			layer = layer,
			n_latent = n_latent,
			spatial_loss = spatial_loss,
			lambda_spatial_loss = lambda_spatial_loss
		)

		# copy the PCA loadings from the RNA model
		with torch.no_grad():
			pca_sp.U.copy_(torch.tensor(adata.varm['PCs']))

		return pca_sp

